#!/usr/local/bin/python3
# -*- coding: utf-8 -*-
#
# archNEMESIS - Python implementation of the NEMESIS radiative transfer and retrieval code
# Variables_0.py - Class to represent the variables (i.e., model parameterisations) in the state vector.
#
# Copyright (C) 2025 Juan Alday, Joseph Penn, Patrick Irwin,
# Jack Dobinson, Jon Mason, Jingxuan Yang
#
# This file is part of archNEMESIS.
#
# archNEMESIS is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.

from __future__ import annotations #  for 3.9 compatability

import os
import os.path
import textwrap
#import sys
from typing import Type, Iterable

import numpy as np
import matplotlib.pyplot as plt

#from archnemesis import *
from archnemesis.Models import Models, ModelBase, ModelParameterEntry
from archnemesis.enums import AtmosphericProfileType, Gas
from archnemesis.helpers import io_helper
from archnemesis.helpers import h5py_helper

import logging
_lgr = logging.getLogger(__name__)
_lgr.setLevel(logging.INFO)

#!/usr/local/bin/python3
# -*- coding: utf-8 -*-

###############################################################################################

"""
Created on Tue Mar 29 17:27:12 2021

@author: juanalday

Model variables Class.
"""

class AprReadError(RuntimeError):
    """
    Something has gone wrong when reading a *.apr file
    """


class Variables_0:

    def __init__(self, NVAR=2, NPARAM=10, NX=10, JPRE=-1, JTAN=-1, JSURF=-1, JALB=-1, JXSC=-1, JRAD=-1, JLOGG=-1, JFRAC=-1):

        """
        Inputs
        ------
        @param NVAR: int,
            Number of model variables to be included
        @param NPARAM: int,
            Number of extra parameters needed to implement the different models       
        @param NX: int,
            Number of points in the state vector
        @param JPRE: int,
            Position of ref. tangent pressure in state vector (if included)
        @param JTAN: int,
            Position of tangent altitude correction in state vector (if included)
        @param JSURF: int,
            Position of surface temperature in state vector (if included)
        @param JALB: int,
            Position of start of surface albedo spectrum in state vector (if included)
        @param JXSC: int,
            Position of start of x-section spectrum in state vector (if included)
        @param JRAD: int,
            Position of radius of the planet in state vector (if included)
        @param JLOGG: int,
            Position of surface log_10(g) of planet in state vector (if included)     
         @param JFRAC: int,
            Position of fractional coverage in state vector (if included)  

        Attributes
        ----------
        @attribute VARIDENT: 2D array (NVAR,3)
            Variable ID
        @attribute VARPARAM: 2D array (NVAR,NPARAM)
            Extra parameters needed to implement the parameterisation
        @attribute NXVAR: 1D array
            Number of points in state vector associated with each variable
        @attribute XA: 1D array
            A priori State vector
        @attribute SA: 2D array
            A priori Covariance matrix of the state vector
        @attribute XN: 1D array
            State vector
        @attribute SX: 2D array
            Covariance matrix of the state vector
        @attribute LX: 1D array
            Flag indicating whether the elements of the state vector are carried in log-scale
        @attribute FIX: 1D array
            Flag indicating whether the elements of the state vector must be fixed
        @attribute NUM: 1D array
            Flag indicating how the gradients with respect to a particular element of the state vector must be computed
            (0) Gradients are computed analytically inside CIRSradg (Atmospheric gradients or Surface temperature) or subspecretg (Others)
            (1) Gradients are computed numerically 
        @attribute DSTEP: 1D array
            For the elements of the state vector whose derivative is being calculated numerically, this array indicates
            the step in the function to be used to calculate the numerical derivative (f' = (f(x+h) - f(x)) / h ). 
            If not specified, this step is assumed to be 5% of the value (h = 0.05*x)

        Methods
        -------
        Variables_0.edit_VARIDENT()
        Variables_0.edit_VARPARAM()
        Variables_0.edit_XA()
        Variables_0.edit_XN()
        Variables_0.edit_LX()
        Variables_0.edit_SA()
        Variables_0.edit_SX()
        Variables_0.calc_NXVAR()
        Variables_0.calc_DSTEP()
        Variables_0.calc_FIX()
        Variables_0.read_hdf5()
        Variables_0.read_apr()
        """

        #Input parameters
        self.NVAR = NVAR
        self.NPARAM = NPARAM
        self.NX = NX
        self.JPRE = JPRE
        self.JTAN = JTAN
        self.JSURF = JSURF
        self.JALB = JALB
        self.JXSC = JXSC
        self.JRAD = JRAD
        self.JLOGG = JLOGG
        self.JFRAC = JFRAC
        
        
        # Input the following profiles using the edit_ methods.
        self.VARIDENT = None # np.zeros(NVAR,3)
        self.VARPARAM = None # np.zeros(NVAR,NPARAM)
        self.NXVAR =  None # np.zeros(NX)
        self.XN = None # np.zeros(NX)
        self.LX = None # np.zeros(NX)
        self.FIX =  None # np.zeros(NX)
        self.SX = None # np.zeros((NX, NX))
        self.NUM = None #np.zeros(NX)
        self.DSTEP = None #np.zeros(NX)
        
        # private attributes
        self._models = None
    ################################################################################################################

    @property
    def models(self) -> list[ModelBase]:
        """
        Returns a list filled with the models whose parameters are in the state vector
        """
        
        if self._models is None:
            raise AttributeError('Models have not been found yet (e.g. by "Variables_0.read_apr(...)")')
        return self._models
    
    @property
    def gas_vmr_models(self) -> Iterable[ModelBase]:
        """
        Returns an iterable of models that deal with gas volume mixing ratios
        """
        return filter(lambda model: hasattr(model, 'target') and model.target == AtmosphericProfileType.GAS_VOLUME_MIXING_RATIO, self.models)
    
    @property
    def aerosol_density_models(self) -> Iterable[ModelBase]:
        """
        Returns an iterable of models that deal with aerosol densities
        """
        return filter(lambda model: hasattr(model, 'target') and model.target == AtmosphericProfileType.AEROSOL_DENSITY, self.models)
    
    
    @property
    def model_parameters(self) -> tuple[dict[str,ModelParameterEntry],...]:
        """
        Returns a tuple of dictionaries that contain the values of the parameters of the models in the state vector
        
        
        ## RETURNS ##
        
        all_model_parameters : tuple[dict[str, ModelParameterEntry], ...]
            A tuple of dictionaries that map model parameter names to the "ModelParameterEntry" for that parameter.
        """
        return tuple(
            model.get_parameters_from_state_vector(
                self.XA,
                self.XN,
                self.LX,
                self.FIX
            )
            for model in self.models
        )
    
    
    def model_parameters_as_string(
            self,
        ) -> str:
        """
        Returns a string of the model parameters (formatted as a table)
        """
        mp = self.model_parameters
        
        
        # General header for output, not part of the table, but used for explanatory information
        p_hdr = '\n'.join((
            'i - index of model parameter in the state vector',
        )) + '\n'
        
        # Header for the table, i.e. column labels
        p_tbl_hdr = ('i', 'model id', 'parameter name', 'apriori value', 'posterior value', 'apriori error')
        p_tbl_col_widths = [len(x) for x in p_tbl_hdr]
        p_tbl_full_col_width = None
        p_str = []
        for i in range(len(mp)+1):
            if i==0:
                p_str.append([p_tbl_hdr])
                continue
            else:
                p_str.append([])
            first1 = True
            for p_name, p_val in mp[i-1].items():
                is_log = self.LX[p_val.sv_slice] == 1
                apriori_error = np.sqrt(np.diag(self.SA[p_val.sv_slice,p_val.sv_slice]))
                apriori_error = np.where(is_log, apriori_error*p_val.apriori_value, apriori_error)
                more_than_one_entry = len(p_val.apriori_value) > 1
                for j, (fix, a, b, s) in enumerate(zip(p_val.is_fixed, p_val.apriori_value, p_val.posterior_value, apriori_error)):
                    
                    p_str[i].append((
                        f'{p_val.sv_slice.start+j}',
                        f'{p_val.model_id}' if first1 else '---', 
                        p_name+f'[{j}]' if more_than_one_entry else p_name, 
                        f'{a:07.2E}', 
                        'FIXED' if fix else f'{b:07.2E}',
                        'FIXED' if fix else f'{s:07.2E}'
                    ))
                    first1 = False
                    #first2 = False
                    p_tbl_col_widths = [max(len(_1), _2) for _1,_2 in zip(p_str[i][-1], p_tbl_col_widths)]
        
        tbl_sec_sep = '|-' + ('-|-'.join(('-'*w for w in p_tbl_col_widths))) + '-|'
        p_tbl_full_col_width = len(tbl_sec_sep)
        
        for i in range(len(p_str)):
            for j in range(len(p_str[i])):
                p_str[i][j] = '| ' + (' | '.join((x.ljust(p_tbl_col_widths[k], ' ') for k,x in enumerate(p_str[i][j])))) + ' |'
            p_str[i] = '\n'.join(p_str[i])
        
        p_str = (f'\n{tbl_sec_sep}\n').join(p_str)
        
        return '\n'.join((
            p_hdr,
            '-'*p_tbl_full_col_width,
            p_str,
            '-'*p_tbl_full_col_width,
        ))
    
    
    def plot_model_parameters(
            self, 
            plot_dir : None | str = None, 
            show : bool = False
        ):
        """
        Plots the parameters of the models in the state vector
        
        ARGUMENTS
            plot_dir : None | str = None
                if not 'None', will save the plot to this directory to save the plot to
                this directory with the file name 'model_parameters'
            show : bool = False
                if True will show the plot interactively
        """
        print(self.model_parameters_as_string())
        
        mp = self.model_parameters
        
        _lgr.debug('all model parameters:')
        for x in mp:
            _lgr.debug(f'\t{x}')
        
        n_rows = len(mp) # number of models we have
        n_cols = 10 # this is the 'resolution' of columns
        
        # Work out how much space each parameter of each model needs on its row
        # is in fraction of row used up by a parameter. I.e. if i^th model has two parameters (p1,p2)
        # and p1 has 2 points, and p2 has 3 points, then frac[i] = (0.4, 0.6)
        f_cols = tuple(
            tuple(
                (v.sv_slice.stop - v.sv_slice.start)
                /(
                    sum(
                        (u.sv_slice.stop-u.sv_slice.start for u in x.values())
                    )
                ) for v in x.values()
            ) for x in mp
        )
        _lgr.debug(f'{n_rows=} {n_cols=}')
        _lgr.debug(f'{f_cols=}')
        
        
        subplot_layout = [['' for c in range(n_cols)] for r in range(n_rows)]
        _lgr.debug(f'subplot_layout.shape = ({len(subplot_layout)}, {len(subplot_layout[0])})')
        
        for r, model_params in enumerate(mp):
            c0 = 0
            c = 0
            f_col = f_cols[r]
            for p_idx, (p_name, p_val) in enumerate(model_params.items()):
                last_col = c0 + f_col[p_idx] * n_cols
                model_param_id_str = f'{p_val.model_id}_{r}_{p_idx}'
                _lgr.debug(f'{model_param_id_str}')
                
                while c < last_col:
                    _lgr.debug(f'{r=} {c=}')
                    _lgr.debug(f'before: {subplot_layout[r][c]=}')
                    subplot_layout[r][c] = model_param_id_str
                    _lgr.debug(f'after: {subplot_layout[r][c]=}')
                    c += 1
                
                c0 = last_col
        
        
        _lgr.debug('subplot_layout:')
        for x in subplot_layout:
            _lgr.debug(f'\t{x}')
            
        
        
        
        f = plt.figure(figsize=(8,12))
        
        axes = f.subplot_mosaic(
            subplot_layout, 
            gridspec_kw=dict(
                wspace=0.2*n_cols, 
                hspace=0.1*n_rows
            )
        )
        
        _lgr.debug(f'{len(axes.values())=}')
        _lgr.debug(f'{tuple(axes.keys())=}')
        
        ax_iter = iter(axes.values())
        
        f.suptitle('Model Parameters\n(in same order as varident)', fontsize=10)
        
        is_first_plot=True
        for model_params in mp:
            is_first_col = True
            for pname, pval in model_params.items():
                ax = next(ax_iter)
                x = range(pval.apriori_value.size)
                
                if is_first_col:
                    is_first_col = False
                    ax.set_title(f'{pval.model_id} : {pname}', fontsize=8)
                else:
                    ax.set_title(pname, fontsize=8)
                
                # draw lines
                ax.plot(x, pval.apriori_value, linestyle='-', linewidth=1, marker='none', color='tab:blue', alpha=0.6)
                ax.plot(x, pval.posterior_value, linestyle='-', linewidth=1, marker='none', color='tab:orange', alpha=0.6)
                
                # draw fixed
                temp = np.array(pval.apriori_value)
                temp[~pval.is_fixed] = np.nan
                ax.plot(x, temp, linestyle='none', linewidth=1, marker='x', color='black', alpha=0.6, label='fixed' if is_first_plot else None)
                
                # draw non-fixed
                temp = np.array(pval.apriori_value)
                temp[pval.is_fixed] = np.nan
                ax.plot(x, temp, linestyle='none', linewidth=1, marker='o', color='tab:blue', alpha=0.6, label='apriori' if is_first_plot else None)
                
                temp = np.array(pval.posterior_value)
                temp[pval.is_fixed] = np.nan
                ax.plot(x, temp, linestyle='none', linewidth=1, marker='o', color='tab:orange', alpha=0.6, label='posterior' if is_first_plot else None)
                
                
                # set plot element styling
                ax.tick_params(axis='both', which='both', labelsize=6)
                
                is_first_plot = False
            
            f.legend()
        
        if plot_dir is not None:
            os.makedirs(plot_dir, exist_ok=True)
            plt.savefig(os.path.join(plot_dir, 'model_parameters'))
        
        if show:
            plt.show()
        
        return
    
    
    def edit_VARIDENT(self, VARIDENT_array):
        """
        Edit the Variable IDs
        @param VARIDENT_array: 2D array
            Parameter IDs defining the parameterisation
        """
        VARIDENT_array = np.array(VARIDENT_array)
        #assert len(VARIDENT_array[:,0]) == self.NVAR, 'VARIDENT should have (NVAR,3) elements'
        #assert len(VARIDENT_array[0,:]) == 3, 'VARIDENT should have (NVAR,3) elements'
        self.VARIDENT = VARIDENT_array

    ################################################################################################################

    def edit_VARPARAM(self, VARPARAM_array):
        """
        Edit the extra parameters needed to implement the parameterisations
        @param VARPARAM_array: 2D array
            Extra parameters defining the model
        """
        VARPARAM_array = np.array(VARPARAM_array)
        #assert len(VARPARAM_array[:,0]) == self.NVAR, 'VARPARAM should have (NVAR,NPARAM) elements'
        #assert len(VARPARAM_array[0,:]) == self.NPARAM, 'VARPARAM should have (NVAR,NPARAM) elements'
        self.VARPARAM = VARPARAM_array

    ################################################################################################################

    def edit_XA(self, XA_array):
        """
        Edit the State Vector.
        @param XA_array: 1D array
            Parameters defining the a priori state vector
        """
        XA_array = np.array(XA_array)
        assert len(XA_array) == self.NX, 'XA should have NX elements'
        self.XA = XA_array

    ################################################################################################################

    def edit_XN(self, XN_array):
        """
        Edit the State Vector.
        @param XN_array: 1D array
            Parameters defining the state vector
        """
        XN_array = np.array(XN_array)
        assert len(XN_array) == self.NX, 'XN should have NX elements'
        self.XN = XN_array

    ################################################################################################################

    def edit_LX(self, LX_array):
        """
        Edit the the flag indicating if the elements are in log-scale
        @param LX_array: 1D array
            Flag indicating whether a particular element of the state 
            vector is in log-scale (1) or not (0)
        """
        LX_array = np.array(LX_array,dtype='int32')
        assert len(LX_array) == self.NX, 'LX should have NX elements'
        self.LX = LX_array  

    ################################################################################################################

    def edit_FIX(self, FIX_array):
        """
        Edit the the flag indicating if the elements are to be fixed
        @param FIX_array: 1D array
            Flag indicating whether a particular element of the state 
            vector is fixed (1) or not (0)
        """
        FIX_array = np.array(FIX_array,dtype='int32')
        assert len(FIX_array) == self.NX, 'FIX should have NX elements'
        self.FIX = FIX_array 

    ################################################################################################################

    def edit_SA(self, SA_array):
        """
        Edit the a priori covariance matrix
        @param SA_array: 2D array
            A priori covariance matrix
        """
        SA_array = np.array(SA_array)
        assert len(SA_array[:,0]) == self.NX, 'SA should have (NX,NX) elements'
        assert len(SA_array[0,:]) == self.NX, 'SA should have (NX,NX) elements'
        self.SA = SA_array 

    ################################################################################################################

    def edit_SX(self, SX_array):
        """
        Edit the state vector covariance matrix
        @param SX_array: 2D array
            State vector covariance matrix
        """
        SX_array = np.array(SX_array)
        assert len(SX_array[:,0]) == self.NX, 'SX should have (NX,NX) elements'
        assert len(SX_array[0,:]) == self.NX, 'SX should have (NX,NX) elements'
        self.SX = SX_array 

    ################################################################################################################

    def calc_NXVAR(self, NPRO, nlocations=1):
        """
        Calculate the array defining the number of parameters in the state 
        vector associated with each model
        @param NXVAR_array: 1D array
            Number of parameters in the state vector associated with each model
        """
        
        assert len(self.models) == self.NVAR, "Must have one model per variable"
        
        
        nxvar = np.zeros(len(self.models),dtype='int32')
        assert self.VARIDENT.ndim == 2, "self.VARIDENT must have 2 dimensions"
        
        for i, model in enumerate(self.models):
            nxvar[i] = model.n_state_vector_entries
        
        
        self.NXVAR = nxvar

    ################################################################################################################

    def calc_DSTEP(self):
        """
        Calculate the step size to be used for the calculation of the numerical derivatives
        f' =  ( f(x+h) - f(x) ) / h
        """

        #Generally, we use h = 0.05*x
        dstep = np.zeros(self.NX)
        dstep[:] = self.XN * 0.05

        #Changing the step size for certain parameterisations
        ix = 0
        for i in range(self.NVAR):

            if self.NVAR>1:
                imod = self.VARIDENT[i,2]
                #ipar = self.VARPARAM[i,0]
            else:
                imod = self.VARIDENT[0,2]
                #ipar = self.VARPARAM[0,0]

            if imod == 228:
                
                #V0,C0,C1,C2,P0,P1,P2,P3
                dstep[ix] = 0.5 * np.sqrt( self.SA[ix,ix] )
                dstep[ix+1] = 0.5 * np.sqrt( self.SA[ix+1,ix+1] )
                dstep[ix+2] = 0.5 * np.sqrt( self.SA[ix+2,ix+2] )
                dstep[ix+3] = 0.5 * np.sqrt( self.SA[ix+3,ix+3] )

            ix = ix + self.NXVAR[i]

        self.DSTEP = dstep

    ################################################################################################################

    def calc_FIX(self):
        """
        Check if the fractional error on any of the state vector parameters is so small 
        that it must be kept constant in the retrieval
        @param FIX: 1D array
            Flag indicating the elements of the state vector that need to be fixed
        """

        minferr = 1.0e-6  #minimum fractional error to fix variable.

        ifix = np.zeros(self.NX,dtype='int32')    
        for ix in range(self.NX):
            xa1 = self.XA[ix]
            ea1 = np.sqrt(abs(self.SA[ix,ix]))

            if self.LX[ix]==1:
                xa1 = np.exp(xa1)
                ea1 = xa1*ea1

            ferr = abs(ea1/xa1)
            if ferr<=minferr:
                ifix[ix] = 1
                
        self.FIX = ifix

    ################################################################################################################

    @staticmethod
    def classify_model_type_from_varident(
            varident : np.ndarray[[3],int],
            ngas : int,
            ndust : int
        ) -> tuple[Type, None | AtmosphericProfileType]:
        """
        Works out the type of model (and subtype if applicable) identified by a VARIDENT triplet.
        
        ## ARGUMENTS ##
            
            varident : np.ndarray[[3],int]
                Three integers that identify a model
                
            ngas : int
                The number of gases present in the reference atmosphere
            
            ndust : int
                The number of aerosol species present in the reference atmosphere
            
        ## RETURNS ##
        
            model_classification : tuple[Type, None | AtmosphericProfileType]
                A tuple containing values that classify the model. From broadest scope to narrowest.
                Currently the tuple has the elements (in order):
                    
                    ModelClass : Type
                        A subclass of archnemesis.Models.ModelBase.ModelBase that denotes the broadest
                        classification of the model. This broadly corresponds to the retrieval component
                        that the model interacts with (e.g. Atmosphere_0, Scatter_0, Measurement_0).
                    
                    ParameterisedTarget : None | AtmosphericProfileType
                        The part of the retrieval component that the model parameterises (and therefore
                        alters). This is 'None' when unknown, or an ENUM corresponding to an attribute
                        of the retrieval component that the model parameterises.
        """
        model_classification = None
        if varident[0] == 0:
            model_classification = ( ModelBase, AtmosphericProfileType.TEMPERATURE)
        elif (varident[0] > 0) and int(varident[0]) in iter(Gas):
            model_classification = ( ModelBase, AtmosphericProfileType.GAS_VOLUME_MIXING_RATIO)
        elif (varident[0] < 0) and (-varident[0]) <= ndust:
            model_classification = ( ModelBase, AtmosphericProfileType.AEROSOL_DENSITY)
        elif (varident[0] < 0) and (-varident[0]) == ndust + 1:
            model_classification = ( ModelBase, AtmosphericProfileType.PARA_H2_FRACTION)
        elif (varident[0] < 0) and (-varident[0]) == ndust + 2:
            model_classification = ( ModelBase, AtmosphericProfileType.FRACTIONAL_CLOUD_COVERAGE)
        else:
            # Other models are classified by their ID number
            model_id_parent_classes = Models[varident[2]].__bases__
            assert len(model_id_parent_classes) == 1, "Only support single inheritance of model classes for now"
            model_classification = (model_id_parent_classes[0],None)
        
        return model_classification
    
    ################################################################################################################

    def read_hdf5(self,runname,npro):
        """
        Read the Variables field of the HDF5 file, which contains information about the variables and
        parametrisations that are to be retrieved, as well as their a priori values.
        These parameters are then included in the Variables class.
        
        N.B. In this code, the apriori and retrieved vectors x are usually
        converted to logs, all except for temperature and fractional scale heights
        This is done to reduce instabilities when different parts of the
        vectors and matrices hold vastly different sized properties. e.g.
        cloud x-section and base height.

        @param runname: str
            Name of the Nemesis run
        @param NPRO: int
            Number of altitude levels in the reference atmosphere
        """

        import h5py

        with h5py.File(runname+'.h5','r') as f:
            #Checking if Variables exists
            e = "/Variables" in f
            if e==False:
                raise ValueError('error :: Variables is not defined in HDF5 file')
            else:
                self.NVAR = h5py_helper.retrieve_data(f, 'Scatter/NVAR', np.int32)
            
    ################################################################################################################

    def read_apr(self,runname,npro,ngas,ndust,nlocations=1):
        """
        Read the .apr file, which contains information about the variables and
        parametrisations that are to be retrieved, as well as their a priori values.
        These parameters are then included in the Variables class.
        
        N.B. In this code, the apriori and retrieved vectors x are usually
        converted to logs, all except for temperature and fractional scale heights
        This is done to reduce instabilities when different parts of the
        vectors and matrices hold vastly different sized properties. e.g.
        cloud x-section and base height.

        Inputs
        ---------------

        @param runname: str
            Name of the Nemesis run
        @param npro: int
            Number of altitude levels in the reference atmosphere
        @param ngas: int
            Number of gasses in the reference atmosphere
        @param ndust: int
            Number of aerosol species in the reference atmosphere
            
        Optional inputs
        ----------------
        
        @param NLOCATIONS: int
            Number of locations in the reference atmosphere/surface
        
        """
        
        if self._models is not None:
            _lgr.warning(f'Already have models for {runname}, will overwrite them as we read the *.apr file.')
        self._models = []
            

        #Open file
        with open(runname+'.apr','r') as f:
    
            #Reading header
            s = f.readline().split()
        
            #Reading first line
            s = f.readline().split()
            nvar = int(s[0])
        
            #Initialise some variables
            jsurf = -1
            jalb = -1
            jxsc = -1
            jtan = -1
            jpre = -1
            #jrad = -1
            jlogg = -1
            jfrac = -1
            sxminfac = 0.001
            mparam = 500        #Giving big sizes but they will be re-sized
            mx = 10000
            varident = np.zeros([nvar,3],dtype='int')
            varparam = np.zeros([nvar,mparam])
            lx = np.zeros([mx],dtype='int')
            x0 = np.zeros([mx])
            sx = np.zeros([mx,mx])
            inum = np.zeros([mx],dtype='int')
            
            #Reading data
            i = 0
            ix = 0
            keep_reading = True
            
            while keep_reading:
            
            
                a = f.readline()
                
                
                if a.strip() == '':
                    # If we have read an empty line
                    if i >= nvar:
                        # if we are past the expected number of VARIDENT entries, we are happy and should stop
                        keep_reading = False
                        break
                    else:
                        # if we are not past the expected number of VARIDENT entries, we are sad
                        keep_reading = False
                        raise AprReadError(f'Expected {nvar} entries in {f.name} but only read {i} before encountering an empty line.')
                else:
                    # We have read a non-empty line
                    if i >= nvar:
                        # If we are reading past the expected number of VARIDENT entries, we are concerned
                        _lgr.warning(f'Expected {nvar} entries in {f.name}, but have found an {i+1}^th entry. Increasing number of variables, extending arrays, and continuing to read file...')
                        # Enlarge arrays where required
                        nvar += 1
                        varident = np.concatenate((varident, np.zeros([1,3],dtype='int')), axis=0)
                        varparam = np.concatenate((varparam, np.zeros([1,mparam])), axis=0)
                
                
                
                s = a.strip().split()
                if len(s) < 3:
                    raise AprReadError(f'VARIDENT entry must have at least 3 components, but {i}^th entry "{a}" has {len(s)}')
                
                try:
                    varident[i,:] = tuple(map(int, s[:3]))
                except Exception as e:
                    raise AprReadError('VARIDENT entry MUST be three (3) integers on a single line, other information beyond that is ignored (as a comment)') from e


                found_model_for_varident = False
                
                for model in Models:
                    
                    if model.is_varident_valid(varident[i]):
                        found_model_for_varident = True
                        
                        try:
                            self._models.append(
                                model.from_apr_to_state_vector(
                                    self, 
                                    f, 
                                    varident[i], 
                                    varparam[i], 
                                    ix, 
                                    lx, 
                                    x0, 
                                    sx,
                                    inum, 
                                    npro, 
                                    ngas,
                                    ndust,
                                    nlocations,
                                    runname,
                                    sxminfac
                                )
                            )
                        except Exception as e:
                            raise AprReadError(f'Failed to read {i}^th model entry (with VARIDENT={varident[i]})') from e
                        
                        _lgr.info(f'\nVariables_0 :: read_apr :: varident {varident[i]}. Constructed model "{model.__name__}" (id={model.id})')
                        try:
                            io_helper.OutWidth.push(io_helper.OutWidth.get() - 2)
                            _lgr.info(textwrap.indent(str(self._models[-1].info(lx,x0)), '  '))
                        finally:
                            io_helper.OutWidth.pop()
                        
                        if varident[i][2]==999:  #Retrieval of surface temperature
                            jsurf = ix
                        elif varident[i][2]==666: #Retrieval of pressure at a givent tangent height
                            jpre = ix
                        elif varident[i][2]==777: #Retrieval of tangent height at a given pressure level
                            jtan = ix
                        
                        ix += self._models[-1].n_state_vector_entries

                
                if not found_model_for_varident:
                    raise AprReadError(f'Variables_0 :: read_apr :: no model found for varident {varident[i]}')
                
                i += 1

        nx = ix
        lx1 = np.zeros(nx,dtype='int32')
        inum1 = np.zeros(nx,dtype='int32')
        xa = np.zeros(nx)
        sa = np.zeros([nx,nx])
        lx1[0:nx] = lx[0:nx]
        inum1[0:nx] = inum[0:nx]
        xa[0:nx] = x0[0:nx]
        sa[0:nx,0:nx] = sx[0:nx,0:nx]

        self.NVAR=nvar
        self.NPARAM=mparam
        self.edit_VARIDENT(varident)
        self.edit_VARPARAM(varparam)
        self.calc_NXVAR(npro,nlocations=nlocations)
        self.JPRE, self.JTAN, self.JSURF, self.JALB, self.JXSC, self.JLOGG, self.JFRAC = jpre, jtan, jsurf, jalb, jxsc, jlogg, jfrac
        self.NX = nx
        self.edit_XA(xa)
        self.edit_XN(xa)
        self.edit_SA(sa)
        self.edit_LX(lx1)
        self.NUM = inum1
        self.calc_DSTEP()
        self.calc_FIX()
        
        ################################################################################################################



#!/usr/local/bin/python3
# -*- coding: utf-8 -*-
#
# archNEMESIS - Python implementation of the NEMESIS radiative transfer and retrieval code
# CIA_0.py - Object to represent the collision-induced absorption (CIA) properties of gases.
#
# Copyright (C) 2025 Juan Alday, Joseph Penn, Patrick Irwin,
# Jack Dobinson, Jon Mason, Jingxuan Yang
#
# This file is part of archNEMESIS.
#
# archNEMESIS is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.

from __future__ import annotations #  for 3.9 compatability

from archnemesis.Data.path_data import archnemesis_resolve_path, archnemesis_indirect_path
import numpy as np
import matplotlib.pyplot as plt
import os
from numba import jit

from archnemesis.helpers import h5py_helper
from archnemesis.enums import Gas, ParaH2Ratio

import logging
_lgr = logging.getLogger(__name__)
_lgr.setLevel(logging.DEBUG)

###############################################################################################

"""
Created on Tue Aug 09 17:27:12 2021

@author: juanalday

Collision-Induced Absorption Class.
"""

class CIA_0:

    def __init__(
            self, 
            runname='', 
            INORMAL=ParaH2Ratio.EQUILIBRIUM, 
            NPAIR=9, 
            NT=25, 
            CIADATA=None, 
            CIATABLE='CO2-CO2_HITRAN.h5', 
            NWAVE=1501, 
            NPARA=0, 
            IPAIRG1=[
                Gas.H2, Gas.H2, Gas.H2, Gas.H2, Gas.H2, Gas.N2, Gas.N2, Gas.CH4, Gas.H2], 
            IPAIRG2=[
                Gas.H2, Gas.He, Gas.H2, Gas.He, Gas.N2, Gas.CH4, Gas.N2, Gas.CH4, Gas.CH4], 
            INORMALT=[
                ParaH2Ratio.EQUILIBRIUM,
                ParaH2Ratio.EQUILIBRIUM,
                ParaH2Ratio.NORMAL,
                ParaH2Ratio.NORMAL,
                ParaH2Ratio.EQUILIBRIUM,
                ParaH2Ratio.EQUILIBRIUM,
                ParaH2Ratio.EQUILIBRIUM,
                ParaH2Ratio.EQUILIBRIUM,
                ParaH2Ratio.EQUILIBRIUM
            ]
        ):
        """
        Inputs
        ------
        @param runname: str,
            Name of the Nemesis run
        @param INORMAL: int,
            Flag indicating whether the ortho/para-H2 ratio is in equilibrium (0 for 1:1) or normal (1 for 3:1)
        @param NPAIR: int,
            Number of gaseous pairs listed 
            (Default = 9 in .cia from Fortran NEMESIS : H2-H2 (eqm), H2-He (eqm), H2-H2 (normal), H2-He (normal), H2-N2, N2-CH4, N2-N2, CH4-CH4, H2-CH4)
        @param NT: int,
            Number of temperature levels over which the CIA data is defined 
        @param NWAVE: int,
            Number of spectral points over which the CIA data is defined
        @param NPARA: int,
            Number of para-H2 fractions listed in the CIA table
        @param FRAC: 1D array (NPARA),
            Fraction of para-H2 in the CIA table (only valid for H2-H2 and H2-He pairs)
        @param IPAIRG1: 1D array (NPAIR),
            First gas of each of the listed pairs (e.g., H2-He ; IPAIRG1 = H2 = 39)
        @param IPAIRG2: 1D array (NPAIR),
            Second gas of each of the listed pairs (e.g., H2-He ; IPAIRG2 = He = 40)
        @param INORMALT: 1D array (NPAIR),
            Flag indicating the equilibrium/normal hydrogen listed in the CIA table (only valid for H2-He and H2-H2, for rest of gases it is not used but needs to be defined)
        @param CIADATA: str
            String indicating where the CIA data files are stored (NOTE: Default location is the Data/cia/ directory)
        @param CIATABLE: str
            String indicating the name of the file storing the CIA table

        Attributes
        ----------
        @attribute WAVEN: 1D array
            Wavenumber array (NOTE: ALWAYS IN WAVENUMBER, NOT WAVELENGTH)
        @attribute TEMP: 1D array
            Temperature levels at which the CIA data is defined (K)
        @attribute K_CIA: 3D array
            CIA cross sections for each pair at each wavenumber and temperature level (cm5 molecule-2 ; NOTE: THIS IS DIFFERENT FROM FORTRAN NEMESIS WHERE THEY ARE LISTED IN CM-1 AMAGAT-2)

        Methods
        ----------
        CIA_0.assess()
        CIA_0.summary_info()
        CIA_0.write_hdf5()
        CIA_0.read_hdf5()
        CIA_0.read_cia()
        CIA_0.plot_cia()
        CIA_0.calc_tau_cia()
        CIA_0.locate_INORMAL_pairs()
        CIA_0.read_ciatable()
        CIA_0.read_ciatable_tab()
        CIA_0.write_ciatable_hdf5()
        CIA_0.read_ciatable_hdf5()
        """

        from archnemesis.Data.path_data import archnemesis_path 

        #Input parameters
        self.runname = runname
        #self.INORMAL : ParaH2Ratio = INORMAL
        self.NPAIR = NPAIR
        self.NPARA = NPARA
        self.IPAIRG1 : list[Gas] = IPAIRG1
        self.IPAIRG2 : list[Gas] = IPAIRG2
        self.INORMALT : list[ParaH2Ratio] = INORMALT
        self.NT = NT
        self.NWAVE = NWAVE
        self.FRAC = np.array([0])
        
        if CIADATA is None:
            self.CIADATA = archnemesis_path()+'archnemesis/Data/cia/'
        else:
            self.CIADATA = CIADATA
            
        self.CIATABLE = CIATABLE

        # Input the following profiles using the edit_ methods.
        self.WAVEN = None # np.zeros(NWAVE)
        self.TEMP = None # np.zeros(NT)
        self.K_CIA = None #np.zeros(NPAIR,NPARA,NT,NWAVE)
        
        # private attributes
        self._inormal = None
        
        # set properties
        self.INORMAL = INORMAL
    
    @property
    def INORMAL(self) -> ParaH2Ratio:
        return self._inormal
    
    @INORMAL.setter
    def INORMAL(self, value):
        self._inormal = ParaH2Ratio(value)
        
    ##################################################################################

    def assess(self):
        """
        Assess whether the different variables have the correct dimensions and types
        """

        #Checking some common parameters to all cases
        assert np.issubdtype(type(self.NPAIR), np.integer) == True , \
            'NPAIR must be int'
        assert self.NPAIR > 0 , \
            'NPAIR must be >0'
            

        assert np.issubdtype(type(self.NT), np.integer) == True , \
            'NT must be int'
        assert self.NT > 0 , \
            'NT must be >0'
            
        assert np.issubdtype(type(self.NWAVE), np.integer) == True , \
            'NWAVE must be int'
        assert self.NWAVE > 0 , \
            'NWAVE must be >0'
        
        assert isinstance(self.INORMAL, ParaH2Ratio) , \
            'INORMAL must be ParaH2Ratio'
            
        assert np.issubdtype(type(self.NPARA), np.integer) == True , \
            'NPARA must be int'
        assert self.NPARA >= 0 , \
            'NPARA must be >= 0'
            
        assert len(self.FRAC) == max(self.NPARA,1) , \
            'FRAC must have size (NPARA) or (1)'
            
        assert len(self.IPAIRG1) == self.NPAIR , \
            'IPAIRG1 must have size (NPAIR)'
            
        assert len(self.IPAIRG2) == self.NPAIR , \
            'IPAIRG2 must have size (NPAIR)'
            
        assert len(self.INORMALT) == self.NPAIR , \
            'INORMALT must have size (NPAIR)'
            
        if self.WAVEN is not None:
            assert len(self.WAVEN) == self.NWAVE , \
                'WAVEN must have size (NWAVE)'
            
        if self.TEMP is not None:
            assert len(self.TEMP) == self.NT , \
                'TEMP must have size (NT)'
                
        if self.K_CIA is not None:
            assert self.K_CIA.shape == (self.NPAIR,max(self.NPARA,1),self.NT,self.NWAVE) , \
                'K_CIA must have size (NPAIR,NPARA or 1,NT,NWAVE)'

    ##################################################################################

    def summary_info(self):
        """
        Print summary information about the CIA class
        """
        
        from archnemesis.Data.gas_data import gas_info
        
        _lgr.info(f'Wavelength range ::  {(self.WAVEN.min(),self.WAVEN.max())}')
        _lgr.info(f'Temperature range ::  {(self.TEMP.min(),self.TEMP.max())}')
        _lgr.info(f'Number of CIA pairs ::  {(self.NPAIR)}')
        _lgr.info('Pairs included in CIA class :: ')
        for i in range(self.NPAIR):
            
            gasname1 = gas_info[str(self.IPAIRG1[i])]['name']
            gasname2 = gas_info[str(self.IPAIRG2[i])]['name']

            label = gasname1+'-'+gasname2
            if self.INORMALT[i]== ParaH2Ratio.NORMAL:
                label = label + " ('normal')"
                
            _lgr.info(label)



    ##################################################################################

    def read_hdf5(self,runname):
        """
        Read the information about the CIA class from the inpit HDF5 file
        @param runname: str
            Name of the NEMESIS run
        """
        
        import h5py

        with h5py.File(runname+'.h5','r') as f:
            #Checking if Spectroscopy exists
            e = "/CIA" in f
            if e==False:
                raise ValueError('error :: CIA is not defined in HDF5 file')
            else:
                self.CIADATA = f['CIA/CIADATA'][0].decode('ascii')
                self.CIATABLE = f['CIA/CIATABLE'][0].decode('ascii')
                self.INORMAL = h5py_helper.retrieve_data(f, 'CIA/INORMAL', lambda x:  ParaH2Ratio(np.int32(x)))
    
        # Resolve archnemesis path if it has been indirected
        self.CIADATA = archnemesis_resolve_path(self.CIADATA)
        
        #Reading the CIA table from the name specified
        self.read_ciatable(self.CIADATA+self.CIATABLE)
    
    ##################################################################################
        
    def write_hdf5(self,runname):
        """
        Write the information about the CIA class from the inpit HDF5 file
        Since the CIA information is actually stored in look-up tables
        @param runname: str
            Name of the NEMESIS run
        @param ciatable: str
            Name of the CIA table (expected to be stored in the Data/cia/ directory)
        """
        
        import h5py

        with h5py.File(runname+'.h5','a') as f:
            #Checking if Spectroscopy already exists
            if ('/CIA' in f)==True:
                del f['CIA']   #Deleting the Spectroscopy information that was previously written in the file

            grp = f.create_group("CIA")

            #Writing the necessary flags
            dset = h5py_helper.store_data(grp, 'INORMAL', int(self.INORMAL))
            dset.attrs['title'] = "Flag indicating whether the ortho/para-H2 ratio is in equilibrium (0 for 1:1) or normal (1 for 3:1)"
            
            #Write the directory where CIA tables are stored
            dt = h5py.special_dtype(vlen=str)
            CIADATA = ['']*1
            
            # Indirect the archnemesis path so it works on other systems
            CIADATA[0] = archnemesis_indirect_path(self.CIADATA)
            
            dset = h5py_helper.store_data(grp, 'CIADATA', CIADATA,dtype=dt)
            dset.attrs['title'] = "Path to directory where CIA table is stored"
            
            # Convert CIATABLE to HDF5 format if we do not have an HDF5 format of this table
            CIATABLE_str = self.CIATABLE
            if CIATABLE_str.endswith('.tab'):
                CIATABLE_str = CIATABLE_str[:-4]+'.h5'
            
            if not os.path.exists(self.CIADATA+CIATABLE_str):
                self.write_ciatable_hdf5(self.CIADATA+CIATABLE_str)
            
            # Write the name of the CIATABLE
            CIATABLE = [CIATABLE_str]
            dset = h5py_helper.store_data(grp, 'CIATABLE', CIATABLE,dtype=dt)
            dset.attrs['title'] = "Name of the CIA table file"
        
        
    ##################################################################################

    def read_cia(self):
        """
        Read the .cia file
        @param runname: str
            Name of the NEMESIS run
        """
        
        #Reading .cia file
        f = open(self.runname+'.cia','r')
        s = f.readline().split()
        cianame = s[0]
        s = f.readline().split()
        dnu = float(s[0])
        s = f.readline().split()
        NPARA = int(s[0])
        f.close()

        self.CIATABLE=cianame

        self.read_ciatable(self.CIADATA+self.CIATABLE, dnu, NPARA)

    ##################################################################################

    def plot_cia(self,logscale=False):
        """
        Subroutine to make a summary plot of the contents of the .cia file
        """

        from archnemesis.Data.gas_data import gas_info

        fig,ax1 = plt.subplots(1,1,figsize=(9,4))

        #labels = ['H$_2$-H$_2$ w equilibrium ortho/para-H$_2$','He-H$_2$ w equilibrium ortho/para-H$_2$','H$_2$-H$_2$ w normal ortho/para-H$_2$','He-H$_2$ w normal ortho/para-H$_2$','H$_2$-N$_2$','N$_2$-CH$_4$','N$_2$-N$_2$','CH$_4$-CH$_4$','H$_2$-CH$_4$']
        for i in range(self.NPAIR):

            gasname1 = gas_info[str(self.IPAIRG1[i])]['name']
            gasname2 = gas_info[str(self.IPAIRG2[i])]['name']

            label = gasname1+'-'+gasname2
            if self.INORMALT[i]==ParaH2Ratio.NORMAL:
                label = label + " ('normal')"

            iTEMP = np.argmin(np.abs(self.TEMP-296.))
            ax1.plot(self.WAVEN,self.K_CIA[i,iTEMP,:],label=label)

        ax1.legend()
        ax1.set_facecolor('lightgray')
        ax1.set_xlabel('Wavenumber (cm$^{-1}$')
        ax1.set_ylabel('CIA cross section (cm$^{5}$ molec$^{-2}$)')
        ax1.grid()
        if logscale==True:
            ax1.set_yscale('log')
        plt.tight_layout()
        plt.show()
        
    ##################################################################################
        
    def locate_INORMAL_pairs(self):
        """
        Subroutine to locate which pairs in the class are dependent on the para/ortho-H2 ratio (i.e., INORMAL = 0 or 1)
        
        Outputs
        -------
        INORMALD(NPAIR) :: Flag indicating whether the pair depends on the ortho/para-H2 ratio (True if it depends)
        """
        
        #We locate the pairs affected by INORMAL by seeing whether some of the IDs for the pairs are repeated
        #If they are repeated, we make sure that they must be defined one with INORMAL = 0 and one with INORMAL = 1
        
        arr = np.vstack([self.IPAIRG1,self.IPAIRG2])
        _, ind = np.unique(arr, axis=1, return_index=True)
        out = np.zeros(shape=arr.shape[1], dtype=bool)
        out[ind] = True
        
        #if out=False then it means that pair is repeated
        iFalse = np.where(out==False)[0]
        
        outx = [False]*self.NPAIR
        for i in range(len(iFalse)):
            
            for j in range(self.NPAIR):
                
                if((self.IPAIRG1[j]==self.IPAIRG1[iFalse[i]]) & (self.IPAIRG2[j]==self.IPAIRG2[iFalse[i]]) ):
                    outx[j] = True
            
        
        #Making sure there are no repeated cases (i.e., the repeated cases have a different INORMAL flag)    
        arr = np.vstack([self.IPAIRG1,self.IPAIRG2,self.INORMALT])
        _, ind = np.unique(arr, axis=1, return_index=True)
        out2 = np.zeros(shape=arr.shape[1], dtype=bool)
        out2[ind] = True
            
        iFalse = np.where(out2==False)[0]
        if len(iFalse)>0:
            raise ValueError('error in locate_INORMAL_pairs :: It appears that there are repeated pairs with the same INORMAL flag')            
            
        return outx
        
    ##################################################################################
    
    def read_ciatable(self, filename, *args):
        """
        Dispatch to correct read_ciatable_X function
        """
        file_read_successfully = False
        
        if filename.endswith(".tab"):
            self.read_ciatable_tab(filename, *args)
            file_read_successfully = True
        elif filename.endswith(".h5"):
            self.read_ciatable_hdf5(filename)
            file_read_successfully = True
        else:
            # Try opening files in different modes use first that works
            if not file_read_successfully:
                try:
                    self.read_ciatable_hdf5(filename+'.h5')
                    file_read_successfully = True
                except Exception:
                    file_read_successfully = False
            if not file_read_successfully:
                try:
                    self.read_ciatable_tab(filename+'.tab', *args)
                    file_read_successfully = True
                except Exception:
                    file_read_successfully = False
        
        if not file_read_successfully:
            raise RuntimeError(f"Failed to read ciatable '{filename}'")
    
    ##################################################################################
    
    def read_ciatable_tab(self, filename, dnu, NPARA):
        """
        Read the CIA look-up table in .tab format
        """
        
        from scipy.io import FortranFile
        
        try:
            f = FortranFile(filename, 'r')
            
            if NPARA != 0:
                NPAIR = 2
                TEMPS = f.read_reals(dtype='float32')
                FRAC = np.abs(f.read_reals(dtype='float32'))
                K_H2H2 = f.read_reals(dtype='float32')
                K_H2HE = f.read_reals(dtype='float32')
                KCIA_list = np.vstack([K_H2H2, K_H2HE]).reshape((-1,), order='F')
                IPAIRG1 = np.array([Gas.H2, Gas.H2])
                IPAIRG2 = np.array([Gas.H2, Gas.He])
                INORMALT = np.array([ParaH2Ratio.EQUILIBRIUM, ParaH2Ratio.EQUILIBRIUM])
                
                self.FRAC = FRAC
                
            # Reading the actual CIA file
            if NPARA == 0:
                NPAIR = 9  # 9 pairs of collision-induced absorption opacities
                TEMPS = f.read_reals(dtype='float64')
                KCIA_list = f.read_reals(dtype='float32')
                IPAIRG1 = np.array([Gas.H2, Gas.H2, Gas.H2, Gas.H2, Gas.H2, Gas.N2, Gas.N2, Gas.CH4, Gas.H2])
                IPAIRG2 = np.array([Gas.H2, Gas.He, Gas.H2, Gas.He, Gas.N2, Gas.CH4, Gas.N2, Gas.CH4, Gas.CH4])
                INORMALT = np.array([
                    ParaH2Ratio.EQUILIBRIUM,
                    ParaH2Ratio.EQUILIBRIUM,
                    ParaH2Ratio.NORMAL,
                    ParaH2Ratio.NORMAL,
                    ParaH2Ratio.EQUILIBRIUM,
                    ParaH2Ratio.EQUILIBRIUM,
                    ParaH2Ratio.EQUILIBRIUM,
                    ParaH2Ratio.EQUILIBRIUM,
                    ParaH2Ratio.EQUILIBRIUM
                ])
        finally:
            f.close()

        NT = len(TEMPS)
        NWAVE = int(len(KCIA_list) / NT / NPAIR / max(NPARA, 1))
        NU_GRID = np.linspace(0, dnu * (NWAVE - 1), NWAVE)
        K_CIA = np.zeros((NPAIR, max(NPARA, 1), NT, NWAVE))  # NPAIR x NPARA x NT x NWAVE

        index = 0
        for iwn in range(NWAVE):
            for itemp in range(NT):
                for ipara in range(max(NPARA, 1)):
                    for ipair in range(NPAIR):
                        K_CIA[ipair, ipara, itemp, iwn] = KCIA_list[index]
                        index += 1

        # Changing the units of the CIA table (NEMESIS format) from cm-1 amagat-2 to cm5 molecule-2
        AMAGAT = 2.68675E19  # molecule cm-3 (definition of amagat unit)
        K_CIA = K_CIA / (AMAGAT**2.)  # cm5 molecule-2

        self.NWAVE = NWAVE
        self.NT = NT
        self.NPAIR = NPAIR
        self.NPARA = NPARA
        self.IPAIRG1 = IPAIRG1
        self.IPAIRG2 = IPAIRG2
        self.INORMALT = INORMALT
        self.WAVEN = NU_GRID
        self.TEMP = TEMPS
        self.K_CIA = K_CIA
    
    ##################################################################################
    
    def write_ciatable_hdf5(self,filename):
        """
        Write the CIA look-up table in an HDF5 file
        """
        
        import h5py

        #Assessing that all the parameters have the correct type and dimension
        self.assess()
        
        if not filename.endswith('.h5'):
            filename += '.h5'
            
        with h5py.File(filename,'w') as f:
            #Writing the main dimensions
            dset = h5py_helper.store_data(f, 'NPARA', data=self.NPARA)
            dset.attrs['title'] = "Number of para-H2 fractions listed in the CIA table"
            
            dset = h5py_helper.store_data(f, 'NPAIR', data=self.NPAIR)
            dset.attrs['title'] = "Number of CIA pairs included in the look-up table"

            dset = h5py_helper.store_data(f, 'NWAVE', data=self.NWAVE)
            dset.attrs['title'] = "Number of wavenumber points in the look-up table"

            dset = h5py_helper.store_data(f, 'NT', data=self.NT)
            dset.attrs['title'] = "Number of temperatures at which the CIA cross sections are tabulated"
            
            dset = h5py_helper.store_data(f, 'IPAIRG1', data=self.IPAIRG1)
            dset.attrs['title'] = "ID of the first gas of each CIA pair (e.g., N2-CO2; IPAIRG1 = N2 = 22)"
            
            dset = h5py_helper.store_data(f, 'IPAIRG2', data=self.IPAIRG2)
            dset.attrs['title'] = "ID of the second gas of each CIA pair (e.g., N2-CO2; IPAIRG2 = CO2 = 2)"
            
            dset = h5py_helper.store_data(f, 'INORMALT', data=self.INORMALT)
            dset.attrs['title'] = "Flag indicating whether the cross sections correspond to equilibrium or normal hydrogen"
            
            dset = h5py_helper.store_data(f, 'WAVEN', data=self.WAVEN)
            dset.attrs['title'] = "Wavenumber"
            dset.attrs['units'] = "cm-1"
            
            dset = h5py_helper.store_data(f, 'TEMP', data=self.TEMP)
            dset.attrs['title'] = "Temperature"
            dset.attrs['units'] = "K"
            
            dset = h5py_helper.store_data(f, 'FRAC', data=self.FRAC)
            dset.attrs['title'] = "Para-H2 fractions"
            
            dset = h5py_helper.store_data(f, 'K_CIA', data=self.K_CIA)
            dset.attrs['title'] = "CIA cross sections"
            dset.attrs['units'] = "cm5 molecule-2"


        
        
    ##################################################################################
        
    def read_ciatable_hdf5(self,filename):
        """
        Read the CIA look-up table from an HDF5 file
        """
        
        import h5py

        if not filename.endswith('.h5'):
            filename += '.h5'
        
        with h5py.File(filename, 'r') as f:
            self.NPARA = np.int32(f.get('NPARA', 0))
            self.NPAIR = h5py_helper.retrieve_data(f, 'NPAIR', np.int32)
            self.NT = h5py_helper.retrieve_data(f, 'NT', np.int32)
            self.NWAVE = h5py_helper.retrieve_data(f, 'NWAVE', np.int32)
                
            self.IPAIRG1 = h5py_helper.retrieve_data(f, 'IPAIRG1', np.array)
            self.IPAIRG2 = h5py_helper.retrieve_data(f, 'IPAIRG2', np.array)
            self.INORMALT = h5py_helper.retrieve_data(f, 'INORMALT', np.array)
            
            self.WAVEN = h5py_helper.retrieve_data(f, 'WAVEN', np.array)
            self.TEMP = h5py_helper.retrieve_data(f, 'TEMP', np.array)
            
            if self.NPARA != 0:
                self.FRAC = h5py_helper.retrieve_data(f, 'FRAC', np.array)
            
            K_CIA = np.zeros((self.NPAIR,max(self.NPARA,1),self.NT,self.NWAVE)) # NPAIR x NPARA x NT x NWAVE
            K_CIA[:,:,:,:] = h5py_helper.retrieve_data(f, 'K_CIA', np.array)
            
            self.K_CIA = K_CIA
        
        self.assess()
        


###############################################################################################

"""
Created on Tue Jul 22 17:27:12 2021

@author: juanalday

Other functions interacting with the CIA class
"""

@jit(nopython=True)
def co2cia(WAVEN):
    """
    Subroutine to return CIA absorption coefficients for CO2-CO2

    @param WAVEN: 1D array
        Wavenumber array (cm-1)
    """

    WAVEL = 1.0e4/WAVEN
    CO2CIA = np.zeros(len(WAVEN))

    #2.3 micron window. Assume de Bergh 1995 a = 4e-8 cm-1/amagat^2
    iin = np.where((WAVEL>=2.15) & (WAVEL<=2.55))
    iin = iin[0]
    if len(iin)>0:
        CO2CIA[iin] = 4.0e-8

    #1.73 micron window. Assume mean a = 6e-9 cm-1/amagat^2
    iin = np.where((WAVEL>=1.7) & (WAVEL<=1.76))
    iin = iin[0]
    if len(iin)>0:
        CO2CIA[iin] = 6.0e-9

    #1.28 micron window. Update from Federova et al. (2014) to
    #aco2 = 1.5e-9 cm-1/amagat^2
    iin = np.where((WAVEL>=1.25) & (WAVEL<=1.35))
    iin = iin[0]
    if len(iin)>0:
        CO2CIA[iin] = 1.5e-9

    #1.18 micron window. Assume a mean a = 1.5e-9 cm-1/amagat^2
    #if(xl.ge.1.05.and.xl.le.1.35)aco2 = 1.5e-9
    #Update from Federova et al. (2014)
    iin = np.where((WAVEL>=1.125) & (WAVEL<=1.225))
    iin = iin[0]
    if len(iin)>0:
        CO2CIA[iin] = 0.5*(0.31+0.79)*1e-9

    #1.10 micron window. Update from Federova et al. (2014)
    iin = np.where((WAVEL>=1.06) & (WAVEL<=1.125))
    iin = iin[0]
    if len(iin)>0:
        CO2CIA[iin] = 0.5*(0.29+0.67)*1e-9
        
    #Changing the units from cm-1 amagat-2 (NEMESIS format) to cm5 molecule-2
    AMAGAT = 2.68675E19 #molecule cm-3 (definition of amagat unit)
    CO2CIA = CO2CIA / (AMAGAT**2.) #cm5 molecule-2

    return CO2CIA

@jit(nopython=True)
def n2n2cia(WAVEN):
    """
    Subroutine to return CIA absorption coefficients for N2-N2
 
    Overtone absorption coef. (km-1 amagat-2) Bob McKellar
    Kindly provided by Caitlin Griffith


    @param WAVEN: 1D array
        Wavenumber array (cm-1)
    """


    WAVEN1 = [4500.0,4505.0,4510.0,4515.0,4520.0,4525.0,4530.0,4535.0,\
    4540.0,4545.0,4550.0,4555.0,4560.0,4565.0,4570.0,4575.0,\
    4580.0,4585.0,4590.0,4595.0,4600.0,4605.0,4610.0,4615.0,\
    4620.0,4625.0,4630.0,4635.0,4640.0,4645.0,4650.0,4655.0,\
    4660.0,4665.0,4670.0,4675.0,4680.0,4685.0,4690.0,4695.0,\
    4700.0,4705.0,4710.0,4715.0,4720.0,4725.0,4730.0,4735.0,\
    4740.0,4745.0,4750.0,4755.0,4760.0,4765.0,4770.0,4775.0,\
    4780.0,4785.0,4790.0,4795.0,4800.0,4805.0,4810.0,4815.0,\
    4820.0,4825.0]
    WAVEN1 = np.array(WAVEN1)

    N2COEF1 = [1.5478185E-05,3.4825567E-05,5.4172953E-05,7.3520343E-05,\
    9.2867725E-05,1.1221511E-04,1.3156250E-04,1.5090988E-04,\
    1.7025726E-04,1.8960465E-04,2.0895203E-04,2.3593617E-04,\
    2.9850862E-04,3.6948317E-04,4.4885988E-04,5.4001610E-04,\
    6.4105232E-04,7.5234997E-04,8.7262847E-04,9.9942752E-04,\
    1.1362602E-03,1.2936132E-03,1.5176521E-03,1.7954395E-03,\
    2.1481151E-03,2.6931590E-03,3.1120952E-03,2.7946872E-03,\
    2.5185575E-03,2.4253442E-03,2.4188559E-03,2.4769977E-03,\
    2.4829037E-03,2.3845681E-03,2.2442993E-03,2.1040305E-03,\
    1.9726211E-03,1.8545000E-03,1.7363789E-03,1.6182578E-03,\
    1.5128252E-03,1.4635258E-03,1.2099572E-03,1.0359654E-03,\
    9.1723543E-04,7.5135247E-04,6.0498451E-04,5.0746030E-04,\
    4.0987082E-04,3.2203691E-04,2.5376283E-04,2.0496233E-04,\
    1.5671484E-04,1.1761552E-04,9.7678370E-05,7.8062728E-05,\
    5.8552457E-05,4.8789554E-05,4.1275161E-05,3.9085765E-05,\
    3.9056369E-05,3.5796973E-05,3.0637581E-05,2.5478185E-05,\
    2.0318790E-05,5.1593952E-06]
    N2COEF1 = np.array(N2COEF1)

    #from scipy.interpolate import interp1d
    #f = interp1d(WAVEN1,N2COEF1)

    #Finding the range within the defined wavenumbers
    N2N2CIA = np.zeros(len(WAVEN))
    iin = np.where((WAVEN>=np.min(WAVEN1)) & (WAVEN<=np.max(WAVEN1)))
    iin = iin[0]

    #N2N2CIA[iin] = f(WAVEN[iin])

    N2N2CIA[iin] = np.interp(WAVEN[iin],WAVEN1,N2COEF1)

    #Convert to cm-1 (amagat)-2
    N2N2CIA = N2N2CIA * 1.0e-5
    
    #Changing the units from cm-1 amagat-2 (NEMESIS format) to cm5 molecule-2
    AMAGAT = 2.68675E19 #molecule cm-3 (definition of amagat unit)
    N2N2CIA = N2N2CIA / (AMAGAT**2.) #cm5 molecule-2

    return N2N2CIA

@jit(nopython=True)
def n2h2cia(WAVEN):
    """
    Subroutine to return CIA absorption coefficients for H2-N2
 
    Absorption coef. (km-1 amagat-2) from McKellar et al.
    Kindly provided by Caitlin Griffith

    @param WAVEN: 1D array
        Wavenumber array (cm-1)
    """

    WAVEN1 = [3995.00,4000.00,4005.00,4010.00,4015.00,4020.00,\
    4025.00,4030.00,4035.00,4040.00,4045.00,4050.00,\
    4055.00,4060.00,4065.00,4070.00,4075.00,4080.00,\
    4085.00,4090.00,4095.00,4100.00,4105.00,4110.00,\
    4115.00,4120.00,4125.00,4130.00,4135.00,4140.00,\
    4145.00,4150.00,4155.00,4160.00,4165.00,4170.00,\
    4175.00,4180.00,4185.00,4190.00,4195.00,4200.00,\
    4205.00,4210.00,4215.00,4220.00,4225.00,4230.00,\
    4235.00,4240.00,4245.00,4250.00,4255.00,4260.00,\
    4265.00,4270.00,4275.00,4280.00,4285.00,4290.00,\
    4295.00,4300.00,4305.00,4310.00,4315.00,4320.00,\
    4325.00,4330.00,4335.00,4340.00,4345.00,4350.00,\
    4355.00,4360.00,4365.00,4370.00,4375.00,4380.00,\
    4385.00,4390.00,4395.00,4400.00,4405.00,4410.00,\
    4415.00,4420.00,4425.00,4430.00,4435.00,4440.00,\
    4445.00,4450.00,4455.00,4460.00,4465.00,4470.00,\
    4475.00,4480.00,4485.00,4490.00,4495.00,4500.00,\
    4505.00,4510.00,4515.00,4520.00,4525.00,4530.00,\
    4535.00,4540.00,4545.00,4550.00,4555.00,4560.00,\
    4565.00,4570.00,4575.00,4580.00,4585.00,4590.00,\
    4595.00,4600.00,4605.00,4610.00,4615.00,4620.00,\
    4625.00,4630.00,4635.00,4640.00,4645.00,4650.00,\
    4655.00,4660.00,4665.00,4670.00,4675.00,4680.00,\
    4685.00,4690.00,4695.00,4700.00,4705.00,4710.00,\
    4715.00,4720.00,4725.00,4730.00,4735.00,4740.00,\
    4745.00,4750.00,4755.00,4760.00,4765.00,4770.00,\
    4775.00,4780.00,4785.00,4790.00,4795.00,4800.00,\
    4805.00,4810.00,4815.00,4820.00,4825.00,4830.00,\
    4835.00,4840.00,4845.00,4850.00,4855.00,4860.00,\
    4865.00,4870.00,4875.00,4880.00,4885.00,4890.00,\
    4895.00,4900.00,4905.00,4910.00,4915.00,4920.00,\
    4925.00,4930.00,4935.00,4940.00,4945.00,4950.00,\
    4955.00,4960.00,4965.00,4970.00,4975.00,4980.00,\
    4985.00,4990.00,4995.00]
    WAVEN1 = np.array(WAVEN1)

    H2N2COEF1 = [3.69231E-04,3.60000E-03,6.83077E-03,1.00615E-02,\
    1.36610E-02,1.84067E-02,2.40000E-02,3.18526E-02,\
    3.97052E-02,4.75578E-02,4.88968E-02,7.44768E-02,\
    9.08708E-02,0.108070,0.139377,0.155680,0.195880,0.228788,\
    0.267880,0.324936,0.367100,0.436444,0.500482,0.577078,\
    0.656174,0.762064,0.853292,0.986708,1.12556,1.22017,\
    1.33110,1.65591,1.69356,1.91446,1.75494,1.63788,\
    1.67026,1.62200,1.60460,1.54774,1.52408,1.48716,\
    1.43510,1.42334,1.34482,1.28970,1.24494,1.16838,\
    1.11038,1.06030,0.977912,0.924116,0.860958,0.807182,\
    0.759858,0.705942,0.680112,0.619298,0.597530,0.550046,\
    0.512880,0.489128,0.454720,0.432634,0.404038,0.378780,\
    0.359632,0.333034,0.317658,0.293554,0.277882,0.262120,\
    0.240452,0.231128,0.210256,0.202584,0.192098,0.181876,\
    0.178396,0.167158,0.171314,0.165576,0.166146,0.170206,\
    0.171386,0.181330,0.188274,0.205804,0.223392,0.253012,\
    0.292670,0.337776,0.413258,0.490366,0.600940,0.726022,\
    0.890254,1.14016,1.21950,1.45480,1.35675,1.53680,\
    1.50765,1.45149,1.38065,1.19780,1.08241,0.977574,\
    0.878010,0.787324,0.708668,0.639210,0.578290,0.524698,\
    0.473266,0.431024,0.392020,0.357620,0.331398,0.299684,\
    0.282366,0.260752,0.242422,0.234518,0.217008,0.212732,\
    0.204464,0.198802,0.199584,0.188652,0.195038,0.191616,\
    0.200324,0.213712,0.224948,0.252292,0.276978,0.318584,\
    0.369182,0.432017,0.527234,0.567386,0.655152,0.660094,\
    0.739228,0.698344,0.662759,0.663277,0.584378,0.535622,\
    0.481566,0.443086,0.400727,0.364086,0.338196,0.303834,\
    0.289236,0.262176,0.247296,0.231594,0.211104,0.205644,\
    0.185118,0.178470,0.170610,0.152406,0.153222,0.132552,\
    0.131400,0.122286,0.109758,0.107472,9.21480E-02,9.09240E-02,\
    8.40520E-02,7.71800E-02,7.03080E-02,6.34360E-02,5.76892E-02,\
    5.32345E-02,4.90027E-02,4.49936E-02,4.12073E-02,3.76437E-02,\
    3.43029E-02,3.11848E-02,2.80457E-02,2.49195E-02,2.19570E-02,\
    1.91581E-02,1.65230E-02,1.40517E-02,1.17440E-02,9.60000E-03,\
    8.40000E-03,7.20000E-03,6.00000E-03,4.80000E-03,3.60000E-03,\
    2.40000E-03,1.20000E-03]
    H2N2COEF1 = np.array(H2N2COEF1)

    #from scipy.interpolate import interp1d
    #f = interp1d(WAVEN1,H2N2COEF1)

    #Finding the range within the defined wavenumbers
    N2H2CIA = np.zeros(len(WAVEN))
    iin = np.where((WAVEN>=np.min(WAVEN1)) & (WAVEN<=np.max(WAVEN1)))[0]

    N2H2CIA[iin] = np.interp(WAVEN[iin],WAVEN1,H2N2COEF1)

    #Convert to cm-1 (amagat)-2
    N2H2CIA = N2H2CIA * 1.0e-5
    
    #Changing the units from cm-1 amagat-2 (NEMESIS format) to cm5 molecule-2
    AMAGAT = 2.68675E19 #molecule cm-3 (definition of amagat unit)
    N2H2CIA = N2H2CIA / (AMAGAT**2.) #cm5 molecule-2

    return N2H2CIA

###############################################################################################

def read_cia_hitran_file(filename):
    """
    Subroutine to read the CIA cross sections from a file written in the HITRAN CIA format
 
    Inputs
    --------
    @param filename: str
        Name of the file
        
    Outputs
    ---------
    @param gasID1: int
        First gas of the pair (e.g., CO2-O2) ; gasID1 = CO2 = 2
    @param gasID2: int
        Second gas of the pair (e.g., CO2-O2) ; gasID2 = O2 = 7
    @param ncases: int
        Number of cases (temperature or spectral ranges) tabulated in the file
    @param temp(ncases): 1D array
        Temperature for each of the cases
    @param nwave(ncases): 1D array
        Number of wavenumbers for each of the cases
    @param wave(nwave,ncases) :: 2D array
        Wavenumber array for each of the cases (cm-1)
    @param k(nwave,ncases): 2D array
        CIA cross section (cm5 molecule-2)
    """
    
    with open(filename,'r') as file1:
    
        
        temp = []
        nwave = []
        wave = []
        k = []
        
        ix = 0
        while True:
            
                    
            if ix==0:
                #Header line for each case
                line = file1.readline()
                # if line is empty
                # end of file is reached
                if not line:
                    break
        
                il = 0
                #paircase = line[il:il+20]
                il = il + 20
                
                #wavemin = float(line[il:il+10])
                il = il + 10
                #wavemax = float(line[il:il+10])
                il = il + 10
                nwavex = int(line[il:il+7])
                il = il + 7
                tempx = float(line[il:il+7])
                il = il + 7
                #ciamax = float(line[il:il+10])
                il = il + 10
                #dwave = float(line[il:il+6])
                il = il + 6
                #comments = line[il:il+27]
                il = il + 27
                #reference = line[il:il+3]
                il = il + 3
                
                ix = 1
                
            else:
            
                temp.append(tempx)
                nwave.append(nwavex)
            
                #Data with cross sections
                for iwave in range(nwavex):
                    line = file1.readline()
                    vals = line.split()
                    wave.append(float(vals[0]))
                    k.append(float(vals[1]))
                    
                ix = 0


    #Re-shaping arrays
    temp = np.array(temp)
    nwave = np.array(nwave,dtype='int32')
    ncases = len(temp)
    waven = np.zeros((nwave.max(),ncases))
    kn = np.zeros((nwave.max(),ncases))
    
    ix = 0
    for i in range(ncases):
        
        waven[0:nwave[i],i] = wave[ix:ix+nwave[i]]
        kn[0:nwave[i],i] = k[ix:ix+nwave[i]]
        ix = ix + nwave[i]
        
    return ncases,temp,nwave,waven,kn




from __future__ import annotations #  for 3.9 compatability

#!/usr/local/bin/python3
# -*- coding: utf-8 -*-
#
# archNEMESIS - Python implementation of the NEMESIS radiative transfer and retrieval code
# LineData_0.py - Class to store line data for a specific gas and isotope.
#
# Copyright (C) 2025 Juan Alday, Joseph Penn, Patrick Irwin,
# Jack Dobinson, Jon Mason, Jingxuan Yang
#
# This file is part of ans.
#
# archNEMESIS is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.


## Imports
# Standard library
from typing import Any, Callable, TYPE_CHECKING
# Third party
import numpy as np
import scipy as sp
import scipy.integrate
import matplotlib.pyplot as plt
# This package
from archnemesis import Data
#from archnemesis import *
import archnemesis as ans
import archnemesis.enums

#import archnemesis.helpers.maths_helper as maths_helper
from archnemesis.helpers.io_helper import SimpleProgressTracker
import archnemesis.database
import archnemesis.database.line_database.hitran
import archnemesis.database.partition_function_database.hitran
from archnemesis.database.filetypes.lbltable import LblDataTProfilesAtPressure#, LblDataTPGrid
from archnemesis.database.datatypes.wave_point import WavePoint
from archnemesis.database.datatypes.wave_range import WaveRange
from archnemesis.database.datatypes.gas_isotopes import GasIsotopes
from archnemesis.database.datatypes.gas_descriptor import RadtranGasDescriptor
# Logging
import logging
_lgr = logging.getLogger(__name__)
_lgr.setLevel(logging.DEBUG)

if TYPE_CHECKING:
    NWAVE = "Number of wave points"

# TODO: 
# * Account for HITRAN weighting by terrestrial abundances .
#   NOTE: do this in the HITRAN.py file, not here as ideally LineData_0 would give 
#         line strengths and absorption coefficients per unit of gas (e.g. per column density,
#         per gram, per mole, something like that) the exact unit to be worked out later.
# * Natural Broadening
# * Pressure shift
# * Multiple ambient gasses. At the moment can only have one, but should be able to define
#   a gas mixture.
# * Additional lineshapes, have a look at https://www.degruyterbrill.com/document/doi/10.1515/pac-2014-0208/html



class LineData_0:
    """
    Clear class for storing line data.
    """
    def __init__(
            self, 
            ID: int = 1,
            ISO: int = 0,
            ambient_gas=ans.enums.AmbientGas.AIR,
            LINE_DATABASE : None | ans.database.protocols.LineDatabaseProtocol = None,
            PARTITION_FUNCTION_DATABASE : None | ans.database.protocols.PartitionFunctionDatabaseProtocol = None,
    ):
        """
        Class to store line data for a specific gas and isotope.

        Inputs
        ------
        @attribute ID: int
            Radtran Gas ID
        
        @attribute ISO: int
            Radtran Isotope ID for each gas, default 0 for all
            isotopes in terrestrial relative abundance
        
        @attribute ambient_gas: enum
            Name of the ambient gas, default AIR. This is used to
            determine the pressure-broadening coefficients
        
        @attribute LINE_DATABASE: None | ans.database.protocols.LineDatabaseProtocol
            Instance of a class that implements the `ans.database.protocols.LineDatabaseProtocol` protocol.
            If `None` will use HITRAN as backend.
        
        @attribute PARTITION_FUNCTION_DATABASE: None | archnemsis.database.protocols.PartitionFunctionDatabaseProtocol
            Instance of a class that implements the `archnemsis.database.protocols.PartitionFunctionDatabaseProtocol` protocol.
            If `None` will use HITRAN as backend.
        

        Attributes
        ----------
        
        @attribute line_data : None | archnemsis.database.protocols.LineDataProtocol
            An object (normally a numpy record array) that implements the `archnemsis.database.protocols.LineDataProtocol`
            protocol. If `None` the data has not been retrieved from the database yet.
            
            If not `None` will have the following attributes:
                
                NU : np.ndarray[['N_LINES_OF_GAS'],float]
                    Transition wavenumber (cm^{-1})
                
                SW : np.ndarray[['N_LINES_OF_GAS'],float]
                    Transition intensity (weighted by isotopologue abundance) (cm^{-1} / molec_cm^{-2})
                
                A : np.ndarray[['N_LINES_OF_GAS'],float]
                    Einstein-A coeifficient (s^{-1})
                
                GAMMA_AMB : np.ndarray[['N_LINES_OF_GAS'],float]
                    Ambient gas broadening coefficient (cm^{-1} atm^{-1})
                
                N_AMB : np.ndarray[['N_LINES_OF_GAS'],float] 
                    Temperature dependent exponent for `GAMMA_AMB` (NUMBER)
                
                DELTA_AMB : np.ndarray[['N_LINES_OF_GAS'],float] 
                    Ambient gas pressure induced line-shift (cm^{-1} atm^{-1})
                
                GAMMA_SELF : np.ndarray[['N_LINES_OF_GAS'],float] 
                    Self broadening coefficient (cm^{-1} atm^{-1})
                
                ELOWER : np.ndarray[['N_LINES_OF_GAS'],float] 
                    Lower state energy (cm^{-1})
        
        @attribute partition_data : None | archnemsis.database.protocols.PartitionFunctionDataProtocol
            An object (normally a numpy record array) that implements the `archnemsis.database.protocols.PartitionFunctionDataProtocol`
            protocol. If `None` the data has not been retrieved from the database yet.
            
            If not `None` will have the following attributes:
            
                TEMP : np.ndarray[['N_TEMPS_OF_GAS'],float]
                    Temperature of tablulated partition function (Kelvin)
                
                Q : np.ndarray[['N_TEMPS_OF_GAS'],float]
                    Tabulated partition function value
        

        Methods
        -------
        LineData_0.assess()
        LineData_0.fetch_linedata(...)
        LineData_0.fetch_partition_function(...)
        """
        
        self._ambient_gas = ambient_gas
        self._line_database = None
        self._partition_function_database = None
        
        self.ID = ID
        self.ISO = ISO
        self.LINE_DATABASE = LINE_DATABASE
        self.PARTITION_FUNCTION_DATABASE = PARTITION_FUNCTION_DATABASE
        
        self.line_data = None
        self.partition_data = None
        
    ##################################################################################

    @property
    def LINE_DATABASE(self) -> ans.database.protocols.LineDatabaseProtocol:
        if self._line_database is None:
            raise RuntimeError('No line database attached to LineData_0 instance')
        return self._line_database

    @LINE_DATABASE.setter
    def LINE_DATABASE(self, value : ans.database.protocols.LineDatabaseProtocol):
        if value is None:
            db = ans.database.line_database.hitran.HITRAN()
            _lgr.info(f'Using default line database {db}')
            self._line_database = db
        else:
            self._line_database = value
    
    @property
    def PARTITION_FUNCTION_DATABASE(self) -> ans.database.protocols.PartitionFunctionDatabaseProtocol:
        if self._partition_function_database is None:
            raise RuntimeError('No partition function database attached to LineData_0 instance')
        return self._partition_function_database

    @PARTITION_FUNCTION_DATABASE.setter
    def PARTITION_FUNCTION_DATABASE(self, value : ans.database.protocols.PartitionFunctionDatabaseProtocol):
        if value is None:
            db = ans.database.partition_function_database.hitran.HITRAN()
            _lgr.info(f'Using default partition function database {db}')
            self._partition_function_database = db
        else:
            self._partition_function_database = value

    @property
    def ambient_gas(self) -> ans.enums.AmbientGas:
        return self._ambient_gas

    @ambient_gas.setter
    def ambient_gas(self, value : int | ans.enums.AmbientGas):
        self._ambient_gas = ans.enums.AmbientGas(value)
    
    @property
    def gas_isotopes(self) -> GasIsotopes:
        return GasIsotopes(self.ID, self.ISO)

    def __repr__(self) -> str:
        return f'LineData_0(ID={self.ID}, ISO={self.ISO}, ambient_gas={self.ambient_gas}, is_line_data_ready={self.is_line_data_ready()}, is_partition_function_ready={self.is_partition_function_ready()}, database={self._database})'

    ##################################################################################
 
    def assess(self) -> None:
        """
        Assess whether the different variables have the correct dimensions and types
        """

        if not isinstance(self.ID, int):
            raise TypeError(f"ID must be an integer, got {type(self.ID)}")
        if not isinstance(self.ISO, int):
            raise TypeError(f"ISO must be an integer, got {type(self.ISO)}")

        assert self.ambient_gas in ans.enums.AmbientGas, \
            f"ambient_gas must be one of {tuple(ans.enums.AmbientGas)}"

        if self.ID < 1:
            raise ValueError(f"ID must be greater than 0, got {self.ID}")
        if self.ISO < 0:   
            raise ValueError(f"ISO must be greater than or equal to 0, got {self.ISO}")
        
    ###########################################################################################################################
    
    def is_line_data_ready(self) -> bool:
        """
        """
        if self.line_data is None:
            return False
        return True
    
    def fetch_linedata(
            self, 
            vmin : float , 
            vmax : float, 
            wave_unit : ans.enums.WaveUnit = ans.enums.WaveUnit.Wavenumber_cm,
            refresh : bool = False, 
    ) -> None:
        """
        Fetch the line data from the specified database, if `refresh` then get the data even if we already have some.
        NOTE: Does not check that the data we already have is valid for the wavelength range (vmin, vmax), only
        checks if `self.is_line_data_ready()` is True or False.
        
        # ARGUMENTS #
            vmin : float
                Minimum wavenumber to get line data for
            vmax : float
                Maximum wavenumber to get line data for
            wave_unit : ans.enums.WaveUnit = ans.enums.WaveUnit.Wavenumber_cm
                Unit of `vmin` and `vmax`, default is wavenumbers in cm^{-1}.
            refresh : bool = False
                If True will retrieve data from database again, even if data
                is already present. NOTE: this will not neccessarily trigger a download
                as the database may cache data.
        """
        assert vmin < vmax, f'Mimimum wave ({vmin}) must be less than maximum wave ({vmax})'
        
        # Turn wavelength range in to Wavenumbers cm^{-1} for internal use
        wave_range = WaveRange(vmin, vmax, wave_unit).to_unit(ans.enums.WaveUnit.Wavenumber_cm)
        
    
        if refresh or not self.is_line_data_ready():
            self.line_data = self.LINE_DATABASE.get_line_data(
                self.gas_isotopes.as_radtran_gasses(), 
                wave_range,
                self.ambient_gas, 
            )
            _lgr.info(f'Retrieved line data from database {self.LINE_DATABASE}')
        else:
            _lgr.info('Line data already loaded')
        
        for gas_desc,v in self.line_data:
            if v is None:
                _lgr.warning(f'Database does not recognise gas {gas_desc} returned `None` for line_data, it should not be included further calculations. If strange things happen, try not including this isotope.')
        
    ###########################################################################################################################
    
    def is_partition_function_ready(self) -> bool:
        return self.partition_data is not None
    
    def fetch_partition_function(self, refresh : bool = False) -> None:
        """
        Get partition function data, if `refresh` then get the data again even if `self.is_partition_function_ready()` is True.
        
        # ARGUMENTS #
            refresh : bool = False
                If True will retrieve data from database again, even if data
                is already present.
        """
        
        
        if refresh or not self.is_partition_function_ready():
            self.partition_data = self.PARTITION_FUNCTION_DATABASE.get_partition_function_data(
                self.gas_isotopes.as_radtran_gasses()
            )
            _lgr.info(f'Retrieved partition function data from database {self.PARTITION_FUNCTION_DATABASE}')
        else:
            _lgr.info('Partition function data already loaded')
    
    ###########################################################################################################################
    
    def calculate_doppler_width(
            self, 
            temp: float, 
    ) -> dict[RadtranGasDescriptor, np.ndarray[['NWAVE'], float]]:
        """
        Calculate Doppler width (HWHM), broadening due to thermal motion.
        NOTE: To get the standard deviation of the gaussian, multiply HWHM by 1/sqrt(2*ln(2))
        
            dlambda/lambda_0 = sqrt(2 ln(2) * (k_b * T)/(m_0 * c^2) )
                             = sqrt(T/m_0) * 1/c * sqrt(2 ln(2) k_b)
                             = sqrt(T/M_0) * 1/c * sqrt(2 ln(2) N_A k_b)
        dlambda - half-width-half-maximum in wavelength space
        lambda_0 - wavelength of line transition
        k_b - boltzmann const
        T - temperature (Kelvin)
        m_0 - mass of a single molecule
        c - speed of light
        M_0 - molecular mass (mass per mole of molecules)
        N_A - avogadro's constant
        
        """
        doppler_width_const_cgs : float  = (1.0 / Data.constants.c_light_cgs) * np.sqrt(2 * np.log(2) * Data.constants.N_avogadro * Data.constants.k_boltzmann_cgs)
        _lgr.debug(f'{doppler_width_const_cgs=}')
        dws = dict() # result
        
        for i, gas_desc in enumerate(self.gas_isotopes.as_radtran_gasses()):
            gas_line_data = self.line_data[gas_desc]
            if gas_line_data is None:
                dws[gas_desc] = np.nan
                continue
            
            dws[gas_desc] = doppler_width_const_cgs * gas_line_data.NU * np.sqrt( temp / gas_desc.molecular_mass)
        
        _lgr.debug(dws)
        return dws
    
    def calculate_lorentz_width(
            self, 
            press: float, 
            temp: float,
            amb_frac: float, # fraction of ambient gas
            tref : float = 296,
    ) -> dict[RadtranGasDescriptor, np.ndarray[['NWAVE'], float]]:
        """
        Calculate pressure-broadened width HWHM (half-width-half-maximum) of cauchy-lorentz distribution.
        """
        _lgr.debug(f'{press=} {temp=} {amb_frac=} {tref=}')
        
        lws = dict() # result
        
        tratio = tref/temp
        
        ONCE_FLAG=True
        
        for i, gas_desc in enumerate(self.gas_isotopes.as_radtran_gasses()):
            gas_line_data = self.line_data[gas_desc]
            
            
            if (_lgr.level == logging.DEBUG) and ONCE_FLAG:
                _lgr.debug(f'{gas_line_data.N_AMB=}')
                _lgr.debug(f'{gas_line_data.GAMMA_AMB=}')
                _lgr.debug(f'{gas_line_data.N_SELF=}')
                _lgr.debug(f'{gas_line_data.GAMMA_SELF=}')
                ONCE_FLAG=False
            
            if gas_line_data is None:
                lws[gas_desc] = np.nan
                continue
            
            lws[gas_desc] = (
                (tratio**gas_line_data.N_AMB) * gas_line_data.GAMMA_AMB * amb_frac 
                 + (tratio**gas_line_data.N_SELF) * gas_line_data.GAMMA_SELF *  (1-amb_frac)
            ) * press
            
        
        _lgr.debug(lws)
        return lws
    
    
    def calculate_partition_sums(self,T) -> dict[RadtranGasDescriptor, float]:
        """
        Calculate the partition functions at any arbitrary temperature

        Args:
            T (float): Temperature (K)

        Returns:
            QT : np.ndarray[['N_GAS_ISOTOPES'], float]
                Partition functions for each of the isotopes at temperature T 
        """
        QTs = dict() # result
        for i, gas_desc in enumerate(self.gas_isotopes.as_radtran_gasses()):
            gas_partition_data = self.partition_data[gas_desc]
            if gas_partition_data is None:
                QTs[gas_desc] = np.nan
                continue
            
            QTs[gas_desc] = np.interp(T, gas_partition_data.TEMP, gas_partition_data.Q)
        
        return QTs
        
    ###########################################################################################################################
    
    def calculate_line_strength(self,T,Tref=296.) -> dict[RadtranGasDescriptor, np.ndarray[['NWAVE'], float]]:
        """
        Calculate the line strengths at any arbitrary temperature

        Args:
            T (float): Temperature (K)
            Tref (float) :: Reference temperature at which the line strengths are listed (default 296 K)

        Returns:
            line_strengths : dict[RadtranGasDescriptor, np.ndarray[['N_GAS_LINES'], float]
                Line strengths at temperature T
        """
        
        line_strengths = dict() # result
        
        qT = self.calculate_partition_sums(T)
        qTref = self.calculate_partition_sums(Tref)
        
        _lgr.debug(f'{Data.constants.c2_cgs=}')
        
        
        t_factor = Data.constants.c2_cgs * (T - Tref)/(T*Tref)
        
        for i, gas_desc in enumerate(self.gas_isotopes.as_radtran_gasses()):
            gas_line_data = self.line_data[gas_desc]
            if gas_line_data is None:
                line_strengths[gas_desc] = np.zeros((0,),dtype=float)
                continue
            
            # Partition function ratio
            q_ratio = qTref[gas_desc] / qT[gas_desc]
            
            boltz = np.exp(t_factor*gas_line_data.ELOWER)
            
            stim = ( 1 - np.exp(-Data.constants.c2_cgs*gas_line_data.NU/T)) / ( 1 - np.exp(-Data.constants.c2_cgs*gas_line_data.NU/Tref))
            
            line_strengths[gas_desc] = gas_line_data.SW * q_ratio * boltz * stim
            
            # Take into account abundances here?
        
        return line_strengths
    

    
    def calculate_line_absorption_in_bins(
            self,
            delta_wn_edges : np.ndarray, # wavenumber bin edges difference from line center (cm^{-1}), (2,NWAVE) low, high
            strength : float, # line strength
            alpha_d : float, # line doppler width (gaussinan HWHM)
            gamma_l : float, # line lorentz width (cauchy-lorentz HWHM)
            lineshape_fn : Callable[[np.ndarray, float], np.ndarray], # function that calculates line shape
            line_integral_points_delta_wn : np.ndarray,
            TEST : bool = False,
    ) -> np.ndarray:
        """
        Calculates line absorbtion at `delta_wn` wavenumber difference from line center
        for a `lineshape_fn`. Return absorption coefficient (cm^2 at `delta_wn` cm^{-1})
        """
        #return strength* lineshape_fn(0.5*np.sum(delta_wn_edges,axis=0), alpha_d, gamma_l) # TESTING just using lineshape, no integral
        
        # Get value of absorption coefficient at all integration points
        # do this here so we don't have to repeat the calculation of endpoints.
        ls_at_ip = lineshape_fn(line_integral_points_delta_wn, alpha_d, gamma_l)
        integral_ls_at_ip = sp.integrate.cumulative_simpson(ls_at_ip, x=line_integral_points_delta_wn, initial=0)
        
        # Assume symmetric so perform small fix
        integral_ls_at_ip += (1 - integral_ls_at_ip[-1])/2
        
        wn_edges_int = np.interp(
            delta_wn_edges,
            line_integral_points_delta_wn,
            integral_ls_at_ip,
            left=0,
            right=1
        )
        
        lineshape = (wn_edges_int[1] - wn_edges_int[0]) / (delta_wn_edges[1]-delta_wn_edges[0])
        
        if TEST:
            fix, ax = plt.subplots(1,2, figsize=(12,12), squeeze=False)
            ax = ax.flatten()
        
            delta_wn_midpoints = 0.5*np.sum(delta_wn_edges, axis=0)
            _lgr.debug(f'{strength=}')
            _lgr.debug(f'{alpha_d=}')
            _lgr.debug(f'{gamma_l=}')
            _lgr.debug(f'{delta_wn_edges=}')
            _lgr.debug(f'{wn_edges_int.shape=}')
            _lgr.debug(f'{wn_edges_int=}')
            _lgr.debug(f'{delta_wn_midpoints=}')
            _lgr.debug(f'{(wn_edges_int[1] - wn_edges_int[0])=}')
            _lgr.debug(f'{(delta_wn_edges[1] - delta_wn_edges[0])=}')
            _lgr.debug(f'{lineshape=}')
            
            ax[0].set_title('lineshape and frac of lineshape in bin')
            ax[0].plot(
                line_integral_points_delta_wn,
                ls_at_ip,
                '.-',
                color='tab:blue',
                linewidth=1,
                alpha=0.6,
            )
            
            ax[0].plot(
                delta_wn_midpoints,
                lineshape,
                '.-',
                color='tab:orange',
                linewidth=1,
                alpha=0.6,
            )
            
            """
            for x in delta_wn_edges[0]:
                ax[0].axvline(
                    x, 
                    color='tab:red',
                    linewidth=0.5,
                    alpha=0.2,
                )
            for x in delta_wn_edges[1]:
                ax[0].axvline(
                    x, 
                    color='tab:red',
                    linestyle=':',
                    linewidth=0.5,
                    alpha=0.2,
                )
            """
            ax[0].set_yscale('log')
            
            
            ax[1].set_title('Cumulative prob.')
            ax[1].plot(
                line_integral_points_delta_wn,
                integral_ls_at_ip,
                '.-',
                color='tab:blue',
                linewidth=1,
                alpha=0.6,
            )
            
            ax[1].plot(
                delta_wn_edges[0],
                wn_edges_int[0],
                '.-',
                color='tab:red',
                linewidth=1,
                alpha=0.6,
            )
            
            ax[1].plot(
                delta_wn_edges[1],
                wn_edges_int[1],
                'x-',
                color='tab:red',
                linewidth=1,
                alpha=0.6,
            )
            
            plt.show()
            
            raise RuntimeError("TESTING")
        
        #sys.exit()
        return strength * lineshape# / alpha_d # absorption coefficient (cm^2)
    
    
    def calculate_absorption_in_bins(
            self,
            waves : np.ndarray, # wavenumber midpoints (if 1D with shape [NWAVE], will calculate bin edges from difference between values), wavenumber bin edges (if 2D with shape [2,NWAVE], index 0 are "left" edges, index 1 are "right" edges)
            temp: float, # Kelvin
            press: float, # Atmospheres
            amb_frac: float = 1, # fraction of ambient gas
            wave_unit : ans.enums.WaveUnit = ans.enums.WaveUnit.Wavenumber_cm, # Unit of input `waves`
            lineshape_fn: Callable[[np.ndarray, float, float], np.ndarray] = Data.lineshapes.voigt, #  lineshape function to use, will be evaluated at `n_line_integral_points`
            line_calculation_wavenumber_window: float = 25.0, # cm^{-1}, contribution from lines outside this region should be modelled as continuum absorption (see page 29 of RADTRANS manual).
            tref : float = 296, # Reference temperature (Kelvin). TODO: This should be set by the database used
            line_strength_cutoff : float = 1E-32, # Strength below which a line is ignored.
            n_line_integral_points : int = 51, # Number of points to use in lineshape integral
            wn_bin : float = -1.0, # Wavenumber (cm^{-1}) bin size. If negative, the bin size will be multiplied by the absolute value (when `waves` is 1-dimensional will calculate bin edges and then apply this scaling). If positive bin size will be set to this value.
            TEST : bool = False, # FOR DEBUGGING: Set to true to turn on testing, at the moment will display plots of the lineshape calculation
            TEST_PREDICATE : Callable[[int,float],bool] = lambda line_index, line_wavenumber : (np.abs(1/(2.371 * 1E-4) - line_wavenumber) < 0.001), # FOR DEBUGGING: Callable that must return True for testing to happen. First argument is the current line index, second is the current line wavenumber. This is here so we can choose which line to apply tests to
            ensure_linedata_downloaded : bool = True,
    ) -> dict[RadtranGasDescriptor, np.ndarray]:
        """
        Calculate total absorption coefficient (cm^2) for wavenumbers (cm^{-1}) multiplied by a factor of 1E20. 
        Returns the value for a single molecule at the specified temperature, pressure, and ambient gas fraction.
        
        For details see "applications" section (at bottom) of https://hitran.org/docs/definitions-and-units/
        """
        INFO_ONCE_FLAG = False
        n_line_progress = 1000
        
        abs_coeffs = dict()
        
        # Convert waves to wavenumber (cm^{-1})
        waves = np.array(WavePoint(waves, wave_unit).to_unit(ans.enums.WaveUnit.Wavenumber_cm).value)
        
        
        
        if waves.ndim == 1:
            _lgr.info(f'Input waves ({waves.shape=}) are midpoints, calculating wave bin edges...')
            wave_midpoints = np.array(waves)
            wave_edges = np.empty((2,wave_midpoints.size), dtype=float)
            wave_edges[0,1:] =  (wave_midpoints[1:]+wave_midpoints[:-1])*0.5 # "Left" bin edges are midpoints except for the first one
            wave_edges[1,:-1] =  (wave_midpoints[1:]+wave_midpoints[:-1])*0.5 # "Right" bin edges are midpoints except for the last one
            
            wave_edges[0,0] = 2*wave_midpoints[0] - wave_edges[1,0] # First "Left" bin edge is guessed to be the same distance from midpoint as first "Right" bin edge
            wave_edges[1,-1] = 2*wave_midpoints[-1] - wave_edges[0,-1] # Last "Right" bin edge is guessed to be the same distance from midpoint as last "Left" bin edge
            
        elif waves.ndim == 2:
            if waves.shape[0] != 2:
                raise ValueError(f'`waves` argument must be either 1-dimensional (interpreted as wave-point midpoints) or 2-dimensional (interpreted as wave-point bin edges). If 2-dimensional must have shape (2,NWAVE). Current `waves` has {waves.ndim=}, {waves.shape=}')
            _lgr.info(f'Input waves ({waves.shape=}) are edges, calculating wave bin midpoints...')
            wave_midpoints = 0.5*np.sum(waves, axis=0)
            wave_edges = np.array(waves)
        else:
            raise ValueError(f'`waves` argument must be either 1-dimensional (interpreted as wave-point midpoints) or 2-dimensional (interpreted as wave-point bin edges). If 2-dimensional must have shape (2,NWAVE). Current `waves` has {waves.ndim=}, {waves.shape=}')
        
        assert wave_midpoints.size >= 3, "Must have at least three wavenumbers in wavenumber grid"
        
        # Sort the wavenumber bins so their midpoints are in ascending order
        # NOTE: `wav_sort_idxs` is used later to put the results into the same order
        # as the input `waves` array.
        wav_sort_idxs = np.argsort(wave_midpoints)
        wave_midpoints = wave_midpoints[wav_sort_idxs]
        wave_edges[0] = wave_edges[0,wav_sort_idxs]
        wave_edges[1] = wave_edges[1,wav_sort_idxs]
        
        
        if wn_bin != -1.0:
            _lgr.debug(f'Applying {wn_bin=}')
            if wn_bin > 0:
                wave_edges[0] = wave_midpoints - wn_bin/2
                wave_edges[1] = wave_midpoints + wn_bin/2
            else:
                wave_edges = wave_midpoints + (wave_midpoints - wave_edges)*wn_bin
        else:
            _lgr.debug(f'{wn_bin=}, no need to alter passed/calculated bins')
        
        
        if wave_midpoints.size == 3:
            _lgr.debug(f'{wave_midpoints.size=} [{wave_midpoints[0]:012.6E}, {wave_midpoints[1]:012.6E}, {wave_midpoints[2]:012.6E}]')
            _lgr.debug(f'"Left" edges : [({wave_edges[0,0]:012.6E}, {wave_edges[0,1]:012.6E}, {wave_edges[0,2]:012.6E}]')
            _lgr.debug(f'"Right" edges: [({wave_edges[1,0]:012.6E}, {wave_edges[1,1]:012.6E}, {wave_edges[1,2]:012.6E}]')
        elif wave_midpoints.size == 4:
            _lgr.debug(f'{wave_midpoints.size=} [{wave_midpoints[0]:012.6E}, {wave_midpoints[1]}, {wave_midpoints[2]:012.6E}, {wave_midpoints[3]:012.6E}]')
            _lgr.debug(f'"Left" edges : [({wave_edges[0,0]:012.6E}, {wave_edges[0,1]:012.6E}, {wave_edges[0,2]:012.6E}, {wave_edges[0,3]:012.6E}]')
            _lgr.debug(f'"Right" edges: [({wave_edges[1,0]:012.6E}, {wave_edges[1,1]:012.6E}, {wave_edges[1,2]:012.6E}, {wave_edges[1,3]:012.6E}]')
        else:
            _lgr.debug(f'{wave_midpoints.size=} [{wave_midpoints[0]:012.6E}, ..., {wave_midpoints[wave_midpoints.size//2]:012.6E}, ..., {wave_midpoints[-1]:012.6E}]')
            _lgr.debug(f'"Left" edges : [({wave_edges[0,0]:012.6E}, ..., {wave_edges[0,wave_midpoints.size//2]:012.6E}, ..., {wave_edges[0,-1]:012.6E}]')
            _lgr.debug(f'"Right" edges: [({wave_edges[1,0]:012.6E}, ..., {wave_edges[1,wave_midpoints.size//2]:012.6E}, ..., {wave_edges[1,-1]:012.6E}]')
            
        
        self.fetch_linedata(vmin = wave_edges[0,0], vmax = wave_edges[1,-1], refresh=ensure_linedata_downloaded)
        
        _lgr.info('Getting lineshape integral points...')
        lip_factor = 1/(np.e*line_calculation_wavenumber_window)
        lip_n1 = n_line_integral_points//2
        lip_n2 = n_line_integral_points - (lip_n1 + 1)
        min_step = lip_factor*(1+line_calculation_wavenumber_window/(2*n_line_integral_points))
        line_integral_points_delta_wn = np.zeros(n_line_integral_points, dtype=float)
        line_integral_points_delta_wn[:lip_n1] = -(np.geomspace(min_step, line_calculation_wavenumber_window+lip_factor, num=lip_n1) - lip_factor)[::-1]
        line_integral_points_delta_wn[-lip_n2:] = np.geomspace(min_step, line_calculation_wavenumber_window+lip_factor, num=lip_n2) - lip_factor

        _lgr.debug(f'line_integral_points_delta_wn={line_integral_points_delta_wn}')
        
        strengths = self.calculate_line_strength(temp, tref)
        alpha_ds = self.calculate_doppler_width(temp)
        gamma_ls = self.calculate_lorentz_width(press, temp, amb_frac, tref)
        
        k_total = np.zeros_like(wave_midpoints, dtype=float) # define here and re-use the memory
        
        
        for i, gas_desc in enumerate(self.gas_isotopes.as_radtran_gasses()):
            _lgr.debug(f'Getting absorbtion coefficient for {i}^th gas {gas_desc=} (of {self.gas_isotopes.n_isotopes} gasses)')
            gas_line_data = self.line_data[gas_desc]
        
            if gas_line_data is None:
                abs_coeffs[gas_desc] = np.zeros_like(wave_midpoints, dtype=float) # All zeros for gasses that are not present in database
                continue
            else:
                # need a new array for each gas, need to define it before using as want to place values in in a specific order
                abs_coeffs[gas_desc] = np.zeros_like(wave_midpoints, dtype=float) 
            
            k_total *= 0.0 # reset 
            
        
            strength = strengths[gas_desc]
            alpha_d = alpha_ds[gas_desc]
            gamma_l = gamma_ls[gas_desc]
            
            mask_lines_above_min_wave= gas_line_data.NU >= (wave_edges[0,0] - line_calculation_wavenumber_window)
            mask_lines_below_max_wave = gas_line_data.NU <= (wave_edges[1,-1] - line_calculation_wavenumber_window)
            line_idxs_to_include = np.nonzero(mask_lines_above_min_wave & mask_lines_below_max_wave)
            idx_min = np.min(line_idxs_to_include)
            idx_max = np.max(line_idxs_to_include)
            n_lines = idx_max - idx_min
            _lgr.debug(f'{idx_min=} {idx_max=} {n_lines=}')
            
            progress_tracker = SimpleProgressTracker(n_lines, f"Computing line contributions from {gas_desc}. ", 5, target_logger=_lgr)
            
            for _j, line_idx in enumerate(range(idx_min, idx_max)):
                if _j % n_line_progress == 0:
                    progress_tracker.n = _j
                    progress_tracker.log_at(logging.INFO)
                
                if strength[line_idx] < line_strength_cutoff:
                    continue
            
                line_wn_mask = np.abs(wave_midpoints - gas_line_data.NU[line_idx]) < line_calculation_wavenumber_window
                n_line_wn_mask = np.count_nonzero(line_wn_mask)

                
                if n_line_wn_mask > 0:
                    delta_wn_edges = wave_edges[:, line_wn_mask] - gas_line_data.NU[line_idx]
                    
                    k_total[line_wn_mask] += self.calculate_line_absorption_in_bins(
                        delta_wn_edges,
                        strength[line_idx],
                        alpha_d[line_idx],
                        gamma_l[line_idx],
                        lineshape_fn,
                        line_integral_points_delta_wn,
                        TEST = TEST and TEST_PREDICATE(line_idx, gas_line_data.NU[line_idx])
                    ) 
                
                # Add in continuum absorption here if required
                if not INFO_ONCE_FLAG:
                    _lgr.info('NOTE: Continuum absorbtion is handled in the lineshape calculation for now, if required may want to separate it for efficiency')
                    INFO_ONCE_FLAG = True
            
            # Ensure abs coeff are not less than zero
            k_total[k_total<0] = 0
            
            # put into absorption coefficient dictionary, multiply by 1E20 factor here
            abs_coeffs[gas_desc][wav_sort_idxs] = k_total*1E20
        
        return abs_coeffs
    
    
    def calculate_monochromatic_line_absorption(
            self,
            delta_wn : np.ndarray, # wavenumber difference from line center (cm^{-1}), (NWAVE)
            mask : np.ndarray, # Boolean mask (NWAVE) that determines if absorption coefficient will be calculated from lineshape (True) or from a 1/(delta_wn^2) fit (False).
            wide_mask : np.ndarray, # Boolean mask (NWAVE) that determines if absorption coefficient will be at all.
            strength : float, # line strength
            alpha_d : float, # line doppler width (gaussinan HWHM)
            gamma_l : float, # line lorentz width (cauchy-lorentz HWHM)
            lineshape_fn : Callable[[np.ndarray, float, float], np.ndarray] = Data.lineshapes.voigt, # function that calculates line shape
            line_calculation_wavenumber_window : float = 25.0, # cm^{-1}, contribution from lines outside this region should be modelled as continuum absorption, should be the same as the window used to make `mask`
            out : np.ndarray | None = None, # if not None, will place the result into this array and return it, otherwise will create an array and return it.
    ) -> np.ndarray:
    
        # At `line_calculation_wavenumber_window` the lineshape_fn and 1/x^2 fit should be equal
        # requires that`line_calculation_wavenumber_window` was applied symmetrically around zero.
        constant = lineshape_fn(np.array([line_calculation_wavenumber_window]), alpha_d, gamma_l) * (line_calculation_wavenumber_window**2)
        
        if out is None:
            out = np.empty_like(delta_wn)
        
        out[wide_mask] = strength * constant / (delta_wn[wide_mask]**2) # calculate "continuum" contribution for the wide mask
        out[mask] = strength * lineshape_fn(delta_wn[mask], alpha_d, gamma_l) # overwrite the "center" with lineshape contribution
        return out
    
    
    def calculate_monochromatic_absorption(
            self,
            waves : np.ndarray, # 1D array with shape [NWAVE]
            temp : float, # kelvin
            press : float, # Atmospheres
            amb_frac : float = 1, # fraction of broadening due to ambient gas
            wave_unit : ans.enums.waveUnit = ans.enums.WaveUnit.Wavenumber_cm,  # unit of `waves` argument
            lineshape_fn : Callable[[np.ndarray, float, float], np.ndarray] = Data.lineshapes.voigt, # lineshape function to use
            line_calculation_wavenumber_window: float = 25.0, # cm^{-1}, contribution from lines outside this region should be modelled as continuum absorption (see page 29 of RADTRANS manual).
            tref : float = 296, # Reference temperature (Kelvin). TODO: This should be set by the database used
            line_strength_cutoff : float = 1E-32, # Strength below which a line is ignored.
            ensure_linedata_downloaded : bool = True,
    ) -> dict[RadtranGasDescriptor, np.ndarray]:
        """
        Calculate total absorption coefficient (cm^2) for wavenumbers (cm^{-1}) multiplied by a factor of 1E20. 
        Returns the value for a single molecule at the specified temperature, pressure, and ambient gas fraction.
        
        For details see "applications" section (at bottom) of https://hitran.org/docs/definitions-and-units/
        """
        INFO_ONCE_FLAG = False
        n_line_progress = 1000
        
        abs_coeffs = dict()
        
        # Convert waves to wavenumber (cm^{-1})
        #in_waves = np.array(waves)
        waves = np.array(WavePoint(waves, wave_unit).to_unit(ans.enums.WaveUnit.Wavenumber_cm).value)
        
        # Remember the ordering we got as input
        wav_sort_idxs = np.argsort(waves)
        
        waves = waves[wav_sort_idxs]
        
        # Download data if we need it
        self.fetch_linedata(
            vmin = waves[0]-line_calculation_wavenumber_window, 
            vmax = waves[-1]+line_calculation_wavenumber_window, 
            wave_unit = ans.enums.WaveUnit.Wavenumber_cm,
            refresh=ensure_linedata_downloaded
        )
        
        strengths = self.calculate_line_strength(temp, tref)
        alpha_ds = self.calculate_doppler_width(temp)
        gamma_ls = self.calculate_lorentz_width(press, temp, amb_frac, tref)
        
        # Define arrays here to re-use memory
        delta_wn = np.zeros_like(waves, dtype=float)
        scratch = np.zeros_like(waves, dtype=float)
        mask_leq = np.zeros_like(waves, dtype=bool)
        mask_geq = np.zeros_like(waves, dtype=bool)
        mask = np.zeros_like(waves, dtype=bool)
        wide_mask_leq = np.zeros_like(waves, dtype=bool)
        wide_mask_geq = np.zeros_like(waves, dtype=bool)
        wide_mask = np.zeros_like(waves, dtype=bool)
        k_total = np.zeros_like(waves, dtype=float)
        
        
        for i, gas_desc in enumerate(self.gas_isotopes.as_radtran_gasses()):
            _lgr.debug(f'Getting absorbtion coefficient for {i}^th gas {gas_desc=} (of {self.gas_isotopes.n_isotopes} gasses)')
            gas_line_data = self.line_data[gas_desc]
        
            if gas_line_data is None:
                abs_coeffs[gas_desc] = np.zeros_like(waves, dtype=float) # All zeros for gasses that are not present in database
                continue
            else:
                # need a new array for each gas, need to define it before using as want to place values in in a specific order
                abs_coeffs[gas_desc] = np.zeros_like(waves, dtype=float) 
            
            k_total.fill(0.0) # reset to all zeros
        
            strength = strengths[gas_desc]
            alpha_d = alpha_ds[gas_desc]
            gamma_l = gamma_ls[gas_desc]
            
            line_idxs_to_include = np.nonzero((waves[0] <= gas_line_data.NU) & (gas_line_data.NU <= waves[-1]))
            idx_min = np.min(line_idxs_to_include)
            idx_max = np.max(line_idxs_to_include)
            n_lines = idx_max - idx_min
            _lgr.debug(f'{idx_min=} {idx_max=} {n_lines=}')
            
            progress_tracker = SimpleProgressTracker(n_lines, f"Computing line contributions from {gas_desc}. ", 5, target_logger=_lgr)
            
            for _j, line_idx in enumerate(range(idx_min, idx_max)):
                if _j % n_line_progress == 0:
                    progress_tracker.n = _j
                    progress_tracker.log_at(logging.INFO)
                
                if strength[line_idx] < line_strength_cutoff:
                    continue
            
                scratch.fill(0.0)
                
                np.subtract(waves, gas_line_data.NU[line_idx], out=delta_wn)
                
                np.less_equal(delta_wn, line_calculation_wavenumber_window, out=mask_leq)
                np.greater_equal(delta_wn, -line_calculation_wavenumber_window, out=mask_geq)
                np.logical_and(mask_leq, mask_geq, out=mask)
                
                np.less_equal(delta_wn, 10*line_calculation_wavenumber_window, out=wide_mask_leq)
                np.greater_equal(delta_wn, -10*line_calculation_wavenumber_window, out=wide_mask_geq)
                np.logical_and(mask_leq, mask_geq, out=wide_mask)

                self.calculate_monochromatic_line_absorption(
                    delta_wn,
                    mask, 
                    wide_mask,
                    strength[line_idx],
                    alpha_d[line_idx],
                    gamma_l[line_idx],
                    lineshape_fn,
                    line_calculation_wavenumber_window,
                    out=scratch
                )
                
                k_total = np.add(k_total, scratch, out=k_total)
                
                # Add in continuum absorption here if required
                if not INFO_ONCE_FLAG:
                    _lgr.info('NOTE: Continuum absorbtion is handled in the lineshape calculation for now, if required may want to separate it for efficiency')
                    INFO_ONCE_FLAG = True
            
            # Ensure abs coeff are not less than zero
            k_total[k_total<0] = 0

            # put into absorption coefficient dictionary, multiply by 1E20 factor here
            abs_coeffs[gas_desc][wav_sort_idxs] = k_total*1E20
        
        return abs_coeffs
        
        
    
    def calculate_absorption_at_temp_pressure_profile(
            self,
            waves : np.ndarray,
            temp_profile : np.ndarray, # Kelvin
            pressure_profile : np.ndarray, # Atmospheres
            delta_temp : float = 5, # Kelvin. Temperature differential
            wave_unit : ans.enums.WaveUnit = ans.enums.WaveUnit.Wavelength_um, # Unit of input `waves`
            wn_bin : float = 0.0, # Wavenumber (cm^{-1}) bin size. If zero will perform monochromatic calculation, if -ve will multiply `waves` spacing by absolute value to get bin size, if +ve specifies an exact bin size.
            **kwargs # Other keyword arguments passed through to `self.calculate_absorption(...)`
    ) -> tuple[LblDataTProfilesAtPressure,...]:
        _lgr.critical('`LineData_0.calculate_absorption_at_temp_pressure_profile` is not quite ready for production use yet. Having trouble with the linetable file format.')
        
        # Need to check that `waves` is evenly spaced and monotonically increasing
        if waves.ndim == 1:
            mid_w = waves
            delta_w = np.diff(waves)
        elif (waves.ndim == 2) and (waves.shape[0] == 2):
            mid_w = 0.5*np.sum(waves, axis=0)
            delta_w = waves
        else:
            raise ValueError(f'`waves` must have one of the following shapes: (NWAVE,) or (2,NWAVE). But {waves.shape=}')
        
        assert np.all(delta_w == delta_w[0]), "`wave` must be an evenly spaced grid"
        
        # Are we performing a monochromatic calculation or a binned calculation
        monochromatic_flag = True if (np.abs(wn_bin) <= 1E-30) else False
        
        # If we are performing a monochromatic calculation, `waves` must only have 1 dimension
        if monochromatic_flag:
            assert waves.ndim == 1, "Monochromatic calculation can only be perfomed when `waves` has a single dimension"
        
        # Define the function that will calculate the absorption
        if monochromatic_flag:
            _lgr.info('Calculating monochromatic absorption...')
            calc_absorption_fn = self.calculate_monochromatic_absorption
        else:
            _lgr.info('Calculating absorption in bins...')
            calc_absorption_fn = lambda *args, **kwargs: self.calculate_absorption_in_bins(*args, wn_bin=wn_bin, **kwargs)
        
        # Get the number of wave points or wave bins we are working with
        n_waves = waves.size if waves.ndim == 1 else waves.shape[1]
        
        # Get temperature points
        temp_points = np.empty((temp_profile.size,2), dtype=float)
        temp_points[:,0] = temp_profile - delta_temp
        temp_points[:,1] = temp_profile + delta_temp
        
        # Get gas descriptors
        gas_descs = tuple(self.gas_isotopes.as_radtran_gasses())
        
        # Allocate result array
        k_total = np.zeros((len(gas_descs), n_waves, pressure_profile.size, 2), dtype=float)
        
        progress_tracker = SimpleProgressTracker(temp_points.size, "Computing absorption coefficients of temperature-pressure profile. ", target_logger=_lgr)
        
        for i, press in enumerate(pressure_profile):
            for j, temp in enumerate(temp_points[i]):
                progress_tracker.log_at_and_increment(logging.INFO)
                #_lgr.info(f'Calculating absorption at temperature-pressure profile point ({i},{j}) of ({pressure_profile.size},{temp_points.shape[1]}). Progress: {i*2+j} / {temp_points.size} [{100.0*(i*2+j)/temp_points.size: 6.2f} %]')
                abs_coeffs = calc_absorption_fn(
                    waves, 
                    temp, 
                    press, 
                    wave_unit = wave_unit,
                    **kwargs
                )
                for k, v in abs_coeffs.items():
                    x = gas_descs.index(k)
                    k_total[x, :, i, j] = v
                    
                    """
                    if (temp > 100) and (press > 7):
                        plt.title(f'{press=}\n{temp=}')
                        plt.plot(waves, v)
                        plt.xlim((1.64,1.68))
                        plt.show()
                        sys.exit()
                    """
        
        result = []
        
        if self.ISO == 0: # In this case, we can put all of the absorption coefficients into a single file
            result.append(
                LblDataTProfilesAtPressure(
                    self.ID,
                    self.ISO,
                    wave_unit,
                    mid_w,
                    pressure_profile,
                    temp_points,
                    np.sum(k_total, axis=0) # add absorption coefficients of different gas isotopes for the combined file
                )
            )
        else: # Otherwise we make a separate file for each gas isotope
            for x, gas_desc in enumerate(gas_descs):
                result.append(
                    LblDataTProfilesAtPressure(
                        gas_desc.gas_id,
                        gas_desc.iso_id,
                        wave_unit,
                        mid_w,
                        pressure_profile,
                        temp_points,
                        k_total[x]
                    )
                )
        
        return result
        
        
        
    
    ###########################################################################################################################
    
    def plot_linedata(
            self, 
            smin : float = 1E-32, 
            logscale : bool = True, 
            scatter_style_kw : dict[str,Any] = {},
            ax_style_kw : dict[str,Any] = {},
            legend_style_kw : dict[str,Any] = {},
    ) -> None:
        """
        Create diagnostic plots of the line data.
        
        ## ARGUMENTS ##
        
            smin : float = 1E-32
                Minimum line strength to plot.
                
            logscale : bool = True
                If True, the y-axis will be in logarithmic scale, else will be linear.
            
            scatter_style_kw : dict[str,Any] = {}
                Dictionary to pass to scatter plots that will set style parameters (e.g. `s` for size, `edgecolor`, `linewidth`,...)
                
            ax_style_kw : dict[str,Any] = {},
                Dictionary to pass to axes that will set style parameters (e.g. `facecolor`,...)
                
            legend_style_kw : dict[str,Any] = {},
                Dictionary to pass to legend that will set style parameters (e.g. `fontsize`, `title_fontsize`, ...)
        """
        
        if not self.is_line_data_ready():
            raise RuntimeError(f'No line data ready in {self}')
        
        scatter_style_defaults = dict(
            s = 15,
            edgecolor='black',
            linewidth=-.2
        )
        scatter_style_defaults.update(scatter_style_kw)
        
        ax_style_defaults = dict(
            facecolor='#EEEEEE'
        )
        ax_style_defaults.update(ax_style_kw)
        
        legend_style_defaults = dict(
            fontsize = 10,
            title_fontsize=12
        )
        legend_style_defaults.update(legend_style_kw)
        
        gas_isotopes = GasIsotopes(self.ID, self.ISO)
        
        f, ax_array = plt.subplots(
            gas_isotopes.n_isotopes+1,1, 
            figsize=(12,4*(gas_isotopes.n_isotopes+1)), 
            gridspec_kw={'hspace':0.3},
            squeeze=False
        )
        ax_array = ax_array.flatten()
        
        combined_ax = ax_array[0]
        combined_ax.set_title('Line data coloured by isotopologue')
        
        line_strengths_max = 0
        for i, gas_desc in enumerate(gas_isotopes.as_radtran_gasses()):
            gas_linedata = self.line_data[gas_desc]
            line_strength_mask = gas_linedata.SW >= smin
            wavenumbers = gas_linedata.NU[line_strength_mask]
            line_strengths = gas_linedata.SW[line_strength_mask]
            
            ls_max = line_strengths.max()
            line_strengths_max = ls_max if ls_max > line_strengths_max else line_strengths_max
            
            # Combined plot, all isotopes on one figure, coloured by isotope
            combined_ax.scatter(
                wavenumbers,
                line_strengths,
                label=f'${ans.Data.gas_data.molecule_to_latex(gas_desc.isotope_name)}$ (ID={int(gas_desc.gas_id)}, ISO={gas_desc.iso_id})',
                **scatter_style_defaults
            )
        
        combined_ax.legend(
            loc='upper left', 
            bbox_to_anchor=(1.01, 1.05),  # Shift legend to the right
            title='Isotope', 
            **legend_style_defaults
        )
        
        if logscale:
            combined_ax.set_yscale('log')
            combined_ax.set_ylim(smin, line_strengths_max * 10)
        combined_ax.set_xlabel('Wavenumber (cm$^{-1}$)')
        combined_ax.set_ylabel('Line strength (cm$^{-1}$ / (molec cm$^{-2}$))')
        combined_ax.set(**ax_style_defaults)
    
        for i, gas_desc in enumerate(gas_isotopes.as_radtran_gasses()):
            gas_linedata = self.line_data[gas_desc]
            line_strength_mask = gas_linedata.SW >= smin
            wavenumbers = gas_linedata.NU[line_strength_mask]
            line_strengths = gas_linedata.SW[line_strength_mask]
            
            # Plots for specific isotopes, coloured by lower energy state
            ax = ax_array[i+1]
            
            ax.set_title(f'Line data for ${ans.Data.gas_data.molecule_to_latex(gas_desc.isotope_name)}$ (ID={int(gas_desc.gas_id)}, ISO={gas_desc.iso_id})')
            p1 = ax.scatter(
                wavenumbers,
                line_strengths,
                c = gas_linedata.ELOWER[line_strength_mask],
                cmap = 'turbo',
                vmin = 0,
                **scatter_style_defaults
            )
            
            # Create a colourbar axes on the right side of ax.
            x_pad = 0.01
            x0, y0, w, h = ax.get_position().bounds
            x1 = x0+w+x_pad
            cax = f.add_axes([x1,y0,0.25*(1-x1),h])
            cbar = plt.colorbar(p1, cax=cax)
            cbar.set_label('Lower state energy (cm$^{-1}$)')
        
            if logscale:
                ax.set_yscale('log')
                ax.set_ylim(smin, line_strengths.max() * 10)
            ax.set_xlabel('Wavenumber (cm$^{-1}$)')
            ax.set_ylabel('Line strength (cm$^{-1}$ / (molec cm$^{-2}$))')
            ax.set(**ax_style_defaults)














import pytest  
import archnemesis as ans
import numpy as np
  
def test_lblconv():
    '''
    Test convolution of the spectra with the ILS
    '''
    
    # Creating function to be convolved with the ILS 
    ##########################################################################

    dvx = 0.1
    vwave = np.arange(0, 50.+dvx, dvx)
    nwave = len(vwave)

    # For demonstration, let's define y as a Gaussian "peak" near the middle
    true_center = 35.
    true_fwhm   = 1.0
    true_sigma  = 0.5 * true_fwhm / np.sqrt(np.log(2))
    y = np.exp(-((vwave - true_center)/true_sigma)**2)
    
    dy = np.zeros((nwave,1))
    dy[:,0] = y * 0.01  # Dummy array

    nshapes = 3
    for ishape in range(nshapes):

        #Performing convolution with archNEMESIS
        ##########################################################################

        Measurement = ans.Measurement_0()
        
        Measurement.NGEOM = 1
        Measurement.FWHM = 1.
        Measurement.ISHAPE = ishape
        Measurement.ISPACE = 0

        vconv = vwave.copy()
        nconv = len(vconv)

        Measurement.NGEOM = 1
        Measurement.NCONV = np.array([len(vconv)])
        vconvx = np.zeros((len(vconv),Measurement.NGEOM))
        vconvx[:,0] = vconv
        Measurement.edit_VCONV(vconvx)
        
        #Dummy statements to check that other functions do not break
        #Measurement.edit_MEAS(vconv/vconv)
        #Measurement.edit_ERRMEAS(vconv/vconv)

        Measurement.build_ils()  #Calculating the ILS
        yconv_an = Measurement.lblconv(vwave,y,IGEOM=0)
        yconv_an2,dummy = Measurement.lblconvg(vwave,y,dy,IGEOM=0)
        
        #Performing the convolution with numpy
        ##########################################################################
        
        #Define the same ILS as in archNEMESIS
        vx = Measurement.VFIL[:,0] - Measurement.VCONV[0,0]
        ax = Measurement.AFIL[:,0]

        #Interpolating the ILS to the same grid as the spectrum (required by numpy)
        vwavex = np.arange(vx.min(),vx.max()+dvx,dvx)
        kernel = np.interp(vwavex,vx,ax)
        
        # Normalize kernel area to 1 (required by numpy)
        kernel_sum = kernel.sum()
        if kernel_sum > 0:
            kernel /= kernel_sum
            
        yconv_np = np.convolve(y, kernel,mode='same')

        #Comparing the results
        assert np.allclose(yconv_an[np.where(yconv_an/yconv_an.max()>=1.0e-2)], yconv_np[(yconv_an/yconv_an.max()>=1.0e-2)], rtol=3.0e-2)
        assert np.allclose(yconv_an2[np.where(yconv_an2/yconv_an2.max()>=1.0e-2)], yconv_np[(yconv_an2/yconv_an2.max()>=1.0e-2)], rtol=3.0e-2)



#!/usr/local/bin/python3
# -*- coding: utf-8 -*-

# flake8: noqa
from .gas_data import *
from .aerosol_data import *
from .planet_data import *
from .path_data import *

import archnemesis.Data.constants as constants
import archnemesis.Data.lineshapes as lineshapes